package codeOfValue;

import edu.rice.cs.mint.runtime.Code;

public class DataTypes {
    public static class WrongTypeException extends RuntimeException {
        public static separable <T> T returningThrow() { throw new WrongTypeException(); }
    }
    
    public static abstract class Value {
        public int intValue() { throw new WrongTypeException(); }
        public boolean booleanValue() { throw new WrongTypeException(); }
        public abstract boolean valueEq(Value other);
    }
    
    public static class IntValue extends Value {
        private int _data;
        public IntValue(int data) { _data = data; }
        public int intValue() { return _data; }
        public boolean valueEq(Value other) {
            if (!(other instanceof IntValue)) throw new WrongTypeException();
            return ((IntValue)other).intValue()==intValue();
        }
        public String toString() { return _data+":IntValue"; }
    }
    
    public static class BooleanValue extends Value {
        private boolean _data;
        public BooleanValue(boolean data) { _data = data; }
        public boolean booleanValue() { return _data; }
        public boolean valueEq(Value other) {
            if (!(other instanceof BooleanValue)) throw new WrongTypeException();
            return ((BooleanValue)other).booleanValue()==booleanValue();
        }
        public String toString() { return _data+":BooleanValue"; }
    }
    
    public static interface Exp {
        public Value eval(Env e, FEnv f);
    }
    
    public static class Val implements Exp {
        private Value _value;
        public Val(Value value) {
            _value = value;
        }
        public Value eval(Env e, FEnv f) {
            return _value;
        }
    }
    
    public static class Var implements Exp {
        private String _s;
        public Var(String s) {
            _s = s;
        }
        public Value eval(Env e, FEnv f) {
            return e.get(_s);
        }
    }
    
    public static class App implements Exp {
        private String _s;
        private Exp _body;
        public App(String s, Exp body) {
            _s = s;
            _body = body;
        }
        public Value eval(Env e, FEnv f) {
            return f.get(_s).apply(_body.eval(e,f));
        }
    }
    
    public static abstract class BinOp implements Exp {
        protected Exp _left, _right;
        public BinOp(Exp left, Exp right) {
            _left = left;
            _right = right;
        }
    }
    
    public static class Add extends BinOp {
        public Add(Exp left, Exp right) {
            super(left, right);
        }
        public Value eval(Env e, FEnv f) {
            return new IntValue(_left.eval(e,f).intValue() + _right.eval(e,f).intValue());
        }
    }
    
    public static class Sub extends BinOp {
        public Sub(Exp left, Exp right) {
            super(left, right);
        }
        public Value eval(Env e, FEnv f) {
            return new IntValue(_left.eval(e,f).intValue() - _right.eval(e,f).intValue());
        }
    }
    
    public static class Mul extends BinOp {
        public Mul(Exp left, Exp right) {
            super(left, right);
        }
        public Value eval(Env e, FEnv f) {
            return new IntValue(_left.eval(e,f).intValue() * _right.eval(e,f).intValue());
        }
    }
    
    public static class Div extends BinOp {
        public Div(Exp left, Exp right) {
            super(left, right);
        }
        public Value eval(Env e, FEnv f) {
            return new IntValue(_left.eval(e,f).intValue() / _right.eval(e,f).intValue());
        }
    }
    
    public static class Equals extends BinOp {
        public Equals(Exp left, Exp right) {
            super(left, right);
        }
        public Value eval(Env e, FEnv f) {
            return new BooleanValue(_left.eval(e,f).valueEq(_right.eval(e,f)));
        }
    }
    
    public static class Less extends BinOp {
        public Less(Exp left, Exp right) {
            super(left, right);
        }
        public Value eval(Env e, FEnv f) {
            return new BooleanValue(_left.eval(e,f).intValue() < _right.eval(e,f).intValue());
        }
    }
    
    public static class And extends BinOp {
        public Exp _left, _right;
        public And(Exp left, Exp right) {
            super(left, right);
        }
        public Value eval(Env e, FEnv f) {
            return new BooleanValue(_left.eval(e,f).booleanValue() && _right.eval(e,f).booleanValue());
        }
    }
    
    public static class Not implements Exp {
        public Exp _exp;
        public Not(Exp exp) {
            _exp = exp;
        }
        public Value eval(Env e, FEnv f) {
            return new BooleanValue(!_exp.eval(e,f).booleanValue());
        }
    }
    
    public static class If implements Exp {
        public Exp _test, _conseq, _alt;
        public If(Exp test, Exp conseq, Exp alt) {
            _test = test;
            _conseq = conseq;
            _alt = alt;
        }
        public Value eval(Env e, FEnv f) {
            return ((_test.eval(e,f).booleanValue())?_conseq:_alt).eval(e,f);
        }
    }
    
    public static class EnvironmentException extends RuntimeException { }
    
    // interface to represent Value -> Value functions
    public static interface ValueFun {
        public Value apply(Value param);
    }
    
    // environment; implemented as a function object from String to Value
    public static interface Env {
        public Value get(String y);
    }
    public static final Env env0 = new Env() {
        public Value get(String s) { throw new EnvironmentException(); }
    };
    
    public static Env ext(final Env env, final String x, final Value v) {
        return new Env() {
            public Value get(String y) {
                if (x.equals(y)) return v; else return env.get(y);
            }
        };
    }
    
    // function environment; implemented as a function object from String to ValueFun
    public static interface FEnv {
        public ValueFun get(String s);
    }
    public static final FEnv fenv0 = new FEnv() {
        public ValueFun get(String s) { throw new EnvironmentException(); }
    };
    
    public static FEnv fext(final FEnv fenv, final String x, final ValueFun v) {
        return new FEnv() {
            public ValueFun get(String y) {
                if (x.equals(y)) return v; else return fenv.get(y);
            }
        };
    }
    
    public static class Declaration {
        public String _fun, _param;
        public Exp _body;
        public Declaration(String fun, String param, Exp body) {
            _fun = fun;
            _param = param;
            _body = body;
        }
        public String fun() { return _fun; }
        public String param() { return _param; }
        public Exp body() { return _body; }
    }
    
    public static class Program {
        public Declaration[] _defs;
        public Exp _body;
        public Program(Exp body, Declaration ... defs) {
            _defs = defs;
            _body = body;
        }
        public Declaration[] defs(){ return _defs; }
        public Exp body() { return _body; }
        
        public Value peval(Env env, FEnv fenv) {
            return peval(env,fenv,0);
        }
        
        private Value peval(final Env env, final FEnv fenv, int defIndex) {
            // match p with
            if (_defs.length<=defIndex) {
                //     Program ([],e) -> eval e env fenv
                return _body.eval(env,fenv);
            }
            else {
                //    |Program (Declaration (s1,s2,e1)::tl,e) ->
                //        let rec f x = eval e1 (ext env s2 x) (ext fenv s1 f)
                //        in peval (Program(tl,e)) env (ext fenv s1 f)
                final Declaration d = _defs[defIndex];
                final String dParam = d._param;
                final String dFun = d._fun;
                final Exp dBody = d._body;
                ValueFun f = new ValueFun() {
                    public Value apply(Value x) {
                        return dBody.eval(ext(env, dParam, x), fext(fenv, dFun, this));
                    }
                };
                return peval(env,fext(fenv, dFun, f),defIndex+1);
            }
        }
    }
    
    /*
     let rec eval e env fenv =
     match e with
     Value i -> i
     | Var s -> env s
     | App (s,e2) -> (fenv s)(eval e2 env fenv)
     | Add (e1,e2) -> (eval e1 env fenv)+(eval e2 env fenv)
     | Sub (e1,e2) -> (eval e1 env fenv)-(eval e2 env fenv)
     | Mul (e1,e2) -> (eval e1 env fenv)*(eval e2 env fenv)
     | Div (e1,e2) -> (eval e1 env fenv)/(eval e2 env fenv)
     | Ifz (e1,e2,e3) -> if (eval e1 env fenv)=0
     then (eval e2 env fenv)
     else (eval e3 env fenv)
     
     let rec peval p env fenv=
     match p with
     Program ([],e) -> eval e env fenv
     |Program (Declaration (s1,s2,e1)::tl,e) ->
     let rec f x = eval e1 (ext env s2 x) (ext fenv s1 f)
     in peval (Program(tl,e)) env (ext fenv s1 f)
     
     */
    
    /*
     let termFact = Program 
     ([Declaration 
     ("f","x",
     Ifz(Var "x", Int 1, Mul(Var"x",(App ("f", Sub(Var "x",Int 1))))))], 
     App ("f", Int 10))
     
     (* let rec f x = if x=0 then 0 else
     if (x-1)=0 then 1 else
     (f (x-1)) + (f (x-2))
     in f 20
     *)
     */
    
    public static final parser.LintParser<Exp> PARSER = new parser.LintParser<Exp>(Exp.class)
        .addSubParser(new parser.LintParser<Value>(Value.class));
    
    public static Program termFact = new Program
        (new App("f", new Val(new IntValue(10))),
         new Declaration
             ("f", "x", new If
                  (new Equals(new Var("x"), new Val(new IntValue(0))),
                   new Val(new IntValue(1)),
                   new Mul(new Var("x"),
                           new App
                               ("f",
                                new Sub
                                    (new Var("x"),
                                     new Val(new IntValue(1))))))));
    
    public static Program parsedTermFact = new Program
        (PARSER.parse("(App f (Val (IntValue 10)))"),
         new Declaration
             ("f", "x", PARSER.parse("(If (Equals (Var x) (Val (IntValue 0)))"+
                                     "    (Val (IntValue 1))"+
                                     "    (Mul (Var x)"+
                                     "         (App f (Sub (Var x)"+
                                     "                     (Val (IntValue 1))))))")));
    
    /*
     let termFib = Program
     ([Declaration 
     ("f","x",
     Ifz(Var "x", Int 0, 
     Ifz(Sub(Var "x",Int 1), Int 1,
     Add(App ("f", Sub(Var "x",Int 1)),
     App ("f", Sub(Var "x",Int 2))))))], 
     App ("f", Int 20))
     
     */
    
    public static Program termFib = new Program
        (new App("f", new Val(new IntValue(10))),
         new Declaration("f", "x", new If
                             (new Equals(new Var("x"), new Val(new IntValue(0))),
                              new Val(new IntValue(0)),
                              new If
                                  (new Equals(new Sub
                                                  (new Var("x"),
                                                   new Val(new IntValue(1))),
                                              new Val(new IntValue(0))),
                                   new Val(new IntValue(1)),
                                   new Add
                                       (new App
                                            ("f",
                                             new Sub
                                                 (new Var("x"),
                                                  new Val(new IntValue(1)))),
                                        new App
                                            ("f",
                                             new Sub
                                                 (new Var("x"),
                                                  new Val(new IntValue(2)))))))));
    
    public static Program parsedTermFib = new Program
        (PARSER.parse("(App f (Val (IntValue 10)))"),
         new Declaration
             ("f", "x", PARSER.parse("(If (Equals (Var x) (Val (IntValue 0)))"+
                                     "    (Val (IntValue 0))"+
                                     "    (If (Equals (Sub (Var x) (Val (IntValue 1)))"+
                                     "                (Val (IntValue 0)))"+
                                     "        (Val (IntValue 1))"+
                                     "        (Add (App f (Sub (Var x) (Val (IntValue 1))))"+
                                     "             (App f (Sub (Var x) (Val (IntValue 2)))))))")));
    
    /*
     let rec f x = if x=0 then 1 else x*(f(x-1)) in f 10
     let rec f x = if x=0 then 1 else x*(f(x-1)) in f (10/2)
     */
    
    public static void main(String[] args) {
        Value i = termFact.peval(env0, fenv0);
        System.out.println("fact(10) = "+i);

        i = parsedTermFact.peval(env0, fenv0);
        System.out.println("fact(10) = "+i);
        
        i = termFib.peval(env0, fenv0);
        System.out.println("fib(10) = "+i);
        
        i = parsedTermFib.peval(env0, fenv0);
        System.out.println("fib(10) = "+i);
    }
}
