package codeOfValue;

import edu.rice.cs.mint.runtime.Code;
import edu.rice.cs.mint.runtime.CodeFree;
import java.io.Serializable;

public class StagedDataTypes {
    public static class WrongTypeException extends RuntimeException {
        public separable WrongTypeException() { super(); }
        public separable WrongTypeException(String msg) { super(msg); }
        public static separable <T> T returningThrow() { throw new WrongTypeException(); }
    }
    public static class EnvironmentException extends RuntimeException {
        public static separable <T> T returningThrow() { throw new EnvironmentException(); }
    }

    public static abstract class Value implements CodeFree, Serializable {
        public separable int intValue() { throw new WrongTypeException(); }
        public separable boolean booleanValue() { throw new WrongTypeException(); }
        public separable abstract boolean valueEq(Value other);
    }
    
    public static class IntValue extends Value {
        private int _data;
        public separable IntValue(int data) { _data = data; }
        public separable int intValue() { return _data; }
        public separable boolean valueEq(Value other) {
            if (!(other instanceof IntValue)) throw new WrongTypeException();
            return ((IntValue)other).intValue()==intValue();
        }
        public String toString() { return _data+":IntValue"; }
    }
    
    public static class BooleanValue extends Value {
        private boolean _data;
        public separable BooleanValue(boolean data) { _data = data; }
        public separable boolean booleanValue() { return _data; }
        public separable boolean valueEq(Value other) {
            if (!(other instanceof BooleanValue)) throw new WrongTypeException();
            return ((BooleanValue)other).booleanValue()==booleanValue();
        }
        public String toString() { return _data+":BooleanValue"; }
    }
    
    public static interface Exp {
        public separable Code<Value> eval(Env e, FEnv f);
    }
    
    public static class Val implements Exp {
        private Code<Value> _value;
        public separable Val(final Value value) {
            _value = <| value |>;
        }
        separable Val(Code<Value> value) { _value = value; }
        public separable Code<Value> eval(Env e, FEnv f) {
            return _value;
        }
    }
    
    public static class Var implements Exp {
        private String _s;
        public separable Var(String s) {
            _s = s; // s must have been interned, so we can use == in the fext method below
        }
        public separable Code<Value> eval(Env e, FEnv f) {
            return e.get(_s);
        }
    }
    
    public static class App implements Exp {
        private final String _s;
        private final Exp _body;
        public separable App(String s, Exp body) {
            _s = s; // s must have been interned, so we can use == in the ext method below
            _body = body;
        }
        public separable Code<Value> eval(final Env e, final FEnv f) {
            final String s = _s;
            final Exp body = _body;
            return <| `(f.get(_s)).apply(`(_body.eval(e,f))) |>;
        }
    }
    
    public static abstract class BinOp implements Exp {
        protected Exp _left, _right;
        public separable BinOp(Exp left, Exp right) {
            _left = left;
            _right = right;
        }
    }
    
    public static class Add extends BinOp {
        public separable Add(Exp left, Exp right) { super(left, right); }
        public separable Code<Value> eval(Env e, FEnv f) {
            return <| (Value) new IntValue(`(_left.eval(e,f)).intValue() + `(_right.eval(e,f)).intValue()) |>;
        }
    }
    
    public static class Sub extends BinOp {
        public separable Sub(Exp left, Exp right) { super(left, right); }
        public separable Code<Value> eval(Env e, FEnv f) {
            return <| (Value) new IntValue(`(_left.eval(e,f)).intValue() - `(_right.eval(e,f)).intValue()) |>;
        }
    }
    
    public static class Mul extends BinOp {
        public separable Mul(Exp left, Exp right) { super(left, right); }
        public separable Code<Value> eval(Env e, FEnv f) {
            return <| (Value) new IntValue(`(_left.eval(e,f)).intValue() * `(_right.eval(e,f)).intValue()) |>;
        }
    }
    
    public static class Div extends BinOp {
        public separable Div(Exp left, Exp right) { super(left, right); }
        public separable Code<Value> eval(Env e, FEnv f) {
            return <| (Value) new IntValue(`(_left.eval(e,f)).intValue() / `(_right.eval(e,f)).intValue()) |>;
        }
    }
    
    public static class Equals extends BinOp {
        public separable Equals(Exp left, Exp right) {
            super(left, right);
        }
        public separable Code<Value> eval(Env e, FEnv f) {
            final Code<Value> l = _left.eval(e,f);
            final Code<Value> r = _right.eval(e,f);
            
            return <| (Value) 
                new BooleanValue((`l).valueEq(`r)) |>;
        }
    }
    
    public static class Less extends BinOp {
        public separable Less(Exp left, Exp right) {
            super(left, right);
        }
        public separable Code<Value> eval(Env e, FEnv f) {
            return <| (Value) new BooleanValue(`(_left.eval(e,f)).intValue()<`(_right.eval(e,f)).intValue()) |>;
        }
    }
    
    public static class And extends BinOp {
        public separable And(Exp left, Exp right) {
            super(left, right);
        }
        public separable Code<Value> eval(Env e, FEnv f) {
            return <| (Value) new BooleanValue(`(_left.eval(e,f)).booleanValue() && `(_right.eval(e,f)).booleanValue()) |>;
        }
    }
    
    public static class Not implements Exp {
        private Exp _exp;
        public separable Not(Exp exp) {
            _exp = exp;
        }
        public separable Code<Value> eval(Env e, FEnv f) {
            return <| (Value) new BooleanValue(!`(_exp.eval(e,f)).booleanValue()) |>;
        }
    }

    public static class If implements Exp {
        private Exp _test, _conseq, _alt;
        public separable If(Exp test, Exp conseq, Exp alt) {
            _test = test;
            _conseq = conseq;
            _alt = alt;
        }
        public separable Code<Value> eval(final Env e, final FEnv f) {
            final Exp lfTest = _test; // hack to help Mint compiler
            final Exp lfConseq = _conseq;
            final Exp lfAlt = _alt;
            return <| `(lfTest.eval(e,f)).booleanValue() ? `(lfConseq.eval(e,f)) : `(lfAlt.eval(e,f)) |>;
        }
    }

    // interface to represent Value -> Value functions
    public static interface ValueFun {
        public Value apply(Value param);
    }
        
    // environment; implemented as a function object from String to Code<Value>
    public static interface Env {
        public separable Code<Value> get(String y);
    }
    public static final Env env0 = new Env() {
        public separable Code<Value> get(String s) {
            throw new EnvironmentException();
        }
    };
    
    public static separable Env ext(final Env env,
                                    final String x,
                                    final Code<Value> v) {
        return new Env() {
            public separable Code<Value> get(String y) {
                // error: if (x.equals(y))
                if (x==y)
                    return v;
                else
                    return env.get(y);
            }
        };
    }     
    
    // function environment; implemented as a function object from String to Code<? extends ValueFun>
    public static interface FEnv {
        public separable Code<? extends ValueFun> get(String s);
    }
    public static final FEnv fenv0 = new FEnv() {
        public separable Code<? extends ValueFun> get(String s) {
            throw new EnvironmentException();
        }
    };
    
    public static separable FEnv fext(final FEnv fenv, final String x, final Code<? extends ValueFun> v) {
        return new FEnv() {
            public separable Code<? extends ValueFun> get(String y) {
                // error: if (x.equals(y))
                if (x==y)
                    return v;
                else
                    return fenv.get(y);
            }
        };
    }
    
    public static class Declaration {
        private String _fun, _param;
        private Exp _body;
        public separable Declaration(String fun, String param, Exp body) {
            _fun = fun; // fun must have been interned, so we can use == in the ext method above
            _param = param; // param must have been interned, so we can use == in the ext method above
            _body = body;
        }
        public String fun() { return _fun; }
        public String param() { return _param; }
        public Exp body() { return _body; }
    }
    
    public static class Program {
        private Declaration[] _defs;
        private Exp _body;
        public separable Program(Exp body, Declaration ... defs) {
            _defs = defs;
            _body = body;
        }
        public Declaration[] defs() { return _defs; }
        public Exp body() { return _body; }
        
        public separable Code<Value> peval(Env env, FEnv fenv) {
            return peval(env,fenv,0);
        }
        
        private separable Code<Value> peval(final Env env, final FEnv fenv, int defIndex) {
            // match p with
            if (_defs.length<=defIndex) {
                //     Program ([],e) -> eval e env fenv
                return _body.eval(env,fenv);
            }
            else {
                //    |Program (Declaration (s1,s2,e1)::tl,e) ->
                //        let rec f x = eval e1 (ext env s2 x) (ext fenv s1 f)
                //        in peval (Program(tl,e)) env (ext fenv s1 f)
                final Declaration d = _defs[defIndex];
                final String dParam = d._param;
                final String dFun = d._fun;
                final Exp dBody = d._body;

                return peval(env,fext(fenv, dFun, <|new ValueFun() {
                    public Value apply(final Value x) {
                        final ValueFun fthis = this;
                        return `(dBody.eval(ext(env, dParam, <| x |>),
                                            fext(fenv, dFun, <| fthis |>)));
                    }
                }|>),defIndex+1);
            }
        }
    }
    
    public static separable Code<Value> bindParameter(Exp functionBody, Code<Value> param) {
        return new Program
            (new App("f", new Val(param)),
             new Declaration
                 ("f", "n", functionBody)).peval(env0, fenv0);
    }
    
    public static final parser.LintParser<Exp> PARSER = new parser.LintParser<Exp>(Exp.class)
        .addSubParser(new parser.LintParser<Value>(Value.class));

    public static Program parsedTermFact = new Program
        (PARSER.parse("(App f (Val (IntValue 10)))"),
         new Declaration
             ("f", "x", PARSER.parse("(If (Equals (Var x) (Val (IntValue 0)))"+
                                     "    (Val (IntValue 1))"+
                                     "    (Mul (Var x)"+
                                     "         (App f (Sub (Var x)"+
                                     "                     (Val (IntValue 1))))))")));

    public static Program parsedTermFib = new Program
        (PARSER.parse("(App f (Val (IntValue 10)))"),
         new Declaration
             ("f", "x", PARSER.parse("(If (Equals (Var x) (Val (IntValue 0)))"+
                                     "    (Val (IntValue 0))"+
                                     "    (If (Equals (Sub (Var x) (Val (IntValue 1)))"+
                                     "                (Val (IntValue 0)))"+
                                     "        (Val (IntValue 1))"+
                                     "        (Add (App f (Sub (Var x) (Val (IntValue 1))))"+
                                     "             (App f (Sub (Var x) (Val (IntValue 2)))))))")));
    
    public static void main(String[] args) {
        Code<Value> sv = parsedTermFact.peval(env0, fenv0);
        System.out.println(sv);
        System.out.println("fact(10) = "+sv.run());
        
        sv = parsedTermFib.peval(env0, fenv0);
        System.out.println(sv);
        System.out.println("fib(10) = "+sv.run());
    }
}
