package gadts;

import java.util.concurrent.*;
import java.util.*;

public class ParLoops {
    public static int NUM_THREADS = Runtime.getRuntime().availableProcessors();
    public static volatile ThreadPoolExecutor pool = null;
    public static synchronized ThreadPoolExecutor pool() {
        if (pool==null) {
            pool = new ThreadPoolExecutor(NUM_THREADS, // number of core threads
                                          NUM_THREADS, // maximum number
                                          0, TimeUnit.SECONDS, // keep alive > core
                                          new LinkedBlockingQueue<Runnable>());
        }
        return pool;
    }

    public static class WrongTypeException extends RuntimeException { }
    
    public static class Value {
        public int intValue() { throw new WrongTypeException(); }
        public double doubleValue() { throw new WrongTypeException(); }
        public boolean booleanValue() { throw new WrongTypeException(); }
        public Value[] arrayValue() { throw new WrongTypeException(); }
    }
    
    public static class IntValue extends Value {
        private int _data;
        public IntValue(int data) { _data = data; }
        public int intValue() { return _data; }
        public double doubleValue() { return _data; } // FIXME: do I want to do that here?
        public String toString() { return _data+":IntValue"; }
    }

    public static class DoubleValue extends Value {
        private double _data;
        public DoubleValue(double data) { _data = data; }
        public int intValue() { return (int)_data; } // FIXME: do I want to do that here?
        public double doubleValue() { return _data; }
        public String toString() { return _data+":DoubleValue"; }
    }
    
    public static class BooleanValue extends Value {
        private boolean _data;
        public BooleanValue(boolean data) { _data = data; }
        public boolean booleanValue() { return _data; }
        public String toString() { return _data+":BooleanValue"; }
    }
    
    public static class ArrayValue extends Value {
        private Value[] _data;
        public ArrayValue(Value[] data) { _data = data; }
        public Value[] arrayValue() { return _data; }
        public String toString() { return java.util.Arrays.toString(_data)+":ArrayValue"; }
    }

    public static interface Exp {
        public Value eval(Env e, FEnv f);
        public RWSet rwSet();
    }
    
    public static class Val implements Exp {
        private Value _value;
        public Val(Value value) {
            _value = value;
        }
        public Value eval(Env e, FEnv f) {
            return _value;
        }
        public RWSet rwSet() { return new RWSet(); }
    }
    
    public static class Var implements Exp {
        private String _s;
        public Var(String s) {
            _s = s;
        }
        public Value eval(Env e, FEnv f) {
            return e.get(_s);
        }
        public RWSet rwSet() { return new RWSet(); }
    }
    
    public static class App implements Exp {
        private String _s;
        private Exp _body;
        public App(String s, Exp body) {
            _s = s;
            _body = body;
        }
        public Value eval(Env e, FEnv f) {
            return f.get(_s).apply(_body.eval(e,f));
        }
        public RWSet rwSet() { return new RWSet(true); }
    }
   
    public static abstract class BinOp implements Exp {
        protected Exp _left, _right;
        public BinOp(Exp left, Exp right) {
            _left = left;
            _right = right;
        }
        public Value eval(Env e, FEnv f) {
            return coerce(_left.eval(e,f), _right.eval(e,f));
        }
        public abstract Value perform(Value v1, Value v2);
        
        public Value coerce(Value v1, Value v2) { // FIXME: Is this the right way to do coercion?
            if ((v1 instanceof IntValue) && (v2 instanceof IntValue)) {
                // both arguments are IntValue, result is IntValue too
                return new IntValue(perform(v1, v2).intValue());
            }
            else {
                return perform(v1, v2);
            }
        }
        
        public RWSet rwSet() {
            return _left.rwSet().union(_right.rwSet());
        }        
    }
    
    public static class Add extends BinOp {
        public Add(Exp left, Exp right) {
            super(left, right);
        }
        public Value perform(Value v1, Value v2) {
            return new DoubleValue(v1.doubleValue() + v2.doubleValue());
        }
    }
    
    public static class Sub extends BinOp {
        public Sub(Exp left, Exp right) {
            super(left, right);
        }
        public Value perform(Value v1, Value v2) {
            return new DoubleValue(v1.doubleValue() - v2.doubleValue());
        }
    }
    
    public static class Mul extends BinOp {
        public Mul(Exp left, Exp right) {
            super(left, right);
        }
        public Value perform(Value v1, Value v2) {
            return new DoubleValue(v1.doubleValue() * v2.doubleValue());
        }
    }
    
    public static class Div extends BinOp {
        public Div(Exp left, Exp right) {
            super(left, right);
        }
        public Value perform(Value v1, Value v2) {
            return new DoubleValue(v1.doubleValue() / v2.doubleValue());
        }
    }
    
    public static class CeilLog extends BinOp {
        public CeilLog(Exp left, Exp right) {
            super(left, right);
        }
        public Value perform(Value v1, Value v2) {
            return new DoubleValue((int)Math.ceil(Math.log(v1.doubleValue()) /
                                                  Math.log(v2.doubleValue())));
        }
    }
    
    public static class Pow extends BinOp {
        public Pow(Exp left, Exp right) {
            super(left, right);
        }
        public Value perform(Value v1, Value v2) {
            return new DoubleValue((int)Math.pow(v1.doubleValue(),
                                                 v2.doubleValue()));
        }
    }

    public static class Floor implements Exp {
        public Exp _exp;
        public Floor(Exp exp) {
            _exp = exp;
        }
        public Value eval(Env e, FEnv f) {
            return new IntValue(_exp.eval(e,f).intValue());
        }
        public RWSet rwSet() {
            return _exp.rwSet();
        }
    }    
    
    public static abstract class ComparisonBinOp extends BinOp {
        public ComparisonBinOp(Exp left, Exp right) {
            super(left, right);
        }
        public Value coerce(Value v1, Value v2) { // FIXME: Is this the right way to do coercion?
            // different coercion here
            if ((v1 instanceof DoubleValue) || (v2 instanceof DoubleValue)) {
                // one of the arguments is DoubleValue, treat both as Double; result is not coerced to anything
                return perform(new DoubleValue(v1.doubleValue()), new DoubleValue(v2.doubleValue()));
            }
            else {
                // do not coerce arguments or result
                return perform(v1, v2);
            }
        }
    }
    
    public static class Equals extends ComparisonBinOp {
        public Equals(Exp left, Exp right) {
            super(left, right);
        }
        public Value perform(Value r, Value l) {
            if ((l instanceof IntValue) && (r instanceof IntValue)) {
                return new BooleanValue(l.intValue()==r.intValue());
            }
            else if ((l instanceof DoubleValue) && (r instanceof DoubleValue)) {
                return new BooleanValue(l.doubleValue()==r.doubleValue());
            }
            else if ((l instanceof BooleanValue) && (r instanceof BooleanValue)) {
                return new BooleanValue(l.booleanValue()==r.booleanValue());
            }
            else if ((l instanceof ArrayValue) && (r instanceof ArrayValue)) {
                return new BooleanValue(l.arrayValue()==r.arrayValue());
            }
            else {
                throw new WrongTypeException();
            }
        }
    }
    
    public static class Less extends ComparisonBinOp {
        public Less(Exp left, Exp right) {
            super(left, right);
        }
        public Value perform(Value v1, Value v2) {
            return new BooleanValue(v1.doubleValue() < v2.doubleValue());
        }
    }
    
    public static class And extends BinOp {
        public Exp _left, _right;
        public And(Exp left, Exp right) {
            super(left, right);
        }
        public Value perform(Value v1, Value v2) {
            return new BooleanValue(v1.booleanValue() && v2.booleanValue());
        }
    }
    
    public static class Not implements Exp {
        public Exp _exp;
        public Not(Exp exp) {
            _exp = exp;
        }
        public Value eval(Env e, FEnv f) {
            return new BooleanValue(!_exp.eval(e,f).booleanValue());
        }
        public RWSet rwSet() {
            return _exp.rwSet();
        }
    }
    
    public static class If implements Exp {
        public Exp _test, _conseq, _alt;
        public If(Exp test, Exp conseq, Exp alt) {
            _test = test;
            _conseq = conseq;
            _alt = alt;
        }
        public Value eval(Env e, FEnv f) {
            return ((_test.eval(e,f).booleanValue())?_conseq:_alt).eval(e,f);
        }
        public RWSet rwSet() {
            return _test.rwSet().union(_conseq.rwSet().union(_alt.rwSet()));
        }
    }
    
    public static class Let implements Exp {
        public String _var;
        public Exp _rhs, _body;
        public Let(String var, Exp rhs, Exp body) {
            _var = var;
            _rhs = rhs;
            _body = body;
        }
        public Value eval(Env e, FEnv f) {
            Env newenv = ext(e, _var, _rhs.eval(e, f));
            return _body.eval(newenv, f);
        }
        public RWSet rwSet() {
            if (_rhs instanceof AllocArray) {
                return _rhs.rwSet().union(_body.rwSet()).removeR(_var).removeW(_var);
            }
            else {
                return _rhs.rwSet().union(_body.rwSet().completeIfContains(_var));
            }
        }
    }
    
    public static class AllocArray implements Exp {
        public Exp _size, _elem;
        public AllocArray(Exp size, Exp elem) {
            _size = size;
            _elem = elem;
        }
        public Value eval(Env e, FEnv f) {
            Value [] arr = allocArray(_size.eval(e, f).intValue (),
                                      _elem.eval(e, f));
            return new ArrayValue(arr);
        }
        public RWSet rwSet() {
            return _size.rwSet().union(_elem.rwSet());
        }
    }
    
    public static class AGet implements Exp {
        public Exp _arr, _index;
        public AGet(Exp arr, Exp index) {
            _arr = arr;
            _index = index;
        }
        public Value eval(Env e, FEnv f) {
            return _arr.eval(e, f).arrayValue()[_index.eval(e, f).intValue()];
        }
        public RWSet rwSet() {
            if (_arr instanceof Var) {
                return _index.rwSet().addR(((Var)_arr)._s);
            } else {
                return _index.rwSet().completeR();
            }
        }
    }
    
    public static class ASet implements Exp {
        public Exp _arr, _index, _val;
        public ASet(Exp arr, Exp index, Exp val) {
            _arr = arr;
            _index = index;
            _val = val;
        }
        public Value eval(Env e, FEnv f) {
            _arr.eval(e, f).arrayValue()[_index.eval(e, f).intValue()]
                = _val.eval(e, f);
            return new IntValue(0);
        }
        public RWSet rwSet() {
            if (_arr instanceof Var) {
                return _index.rwSet().union(_val.rwSet()).addW(((Var)_arr)._s);
            } else {
                return _index.rwSet().union(_val.rwSet()).completeW();
            }
        }
    }
    
    public static class For implements Exp {
        String _var;
        public Exp _bound, _body;
        public For(String var, Exp bound, Exp body) {
            _var = var;
            _bound = bound;
            _body = body;
        }
        
        Callable<Void> mk_task(final Env e, final FEnv f, final int lower, final int upper) {
            return new Callable<Void>() {
                public Void call() {
                    for (int j = lower; j < upper; ++j) {
                        _body.eval(ext(e, _var, new IntValue(j)), f);
                    }
                    return null;
                }
            };
        }
        
        public Value eval(final Env e, final FEnv f) {
            LoopRunner runner = new LoopRunner() {
                public int run(int lower, int upper) {
                    // util.IO.addOutput("For "+_var+" runs "+lower+" <= "+_var+" < "+upper);
                    for (int i = lower; i < upper; ++i) {
                        final int fi = i;
                        _body.eval(ext(e, _var, new IntValue(i)), f);
                    }
                    // util.IO.addOutput("For "+_var+" ran "+lower+" <= "+_var+" < "+upper);
                    return 0;
                }};
            int bound = _bound.eval(e, f).intValue();
            if (rwOverlaps(e, _body.rwSet()) || (pool().getMaximumPoolSize()-pool().getActiveCount())<NUM_THREADS) {
                // util.IO.addOutput("For "+_var+" runs in series");
                return new IntValue(runner.runSerial(bound));
            } else {
                // util.IO.addOutput("For "+_var+" runs in parallel");
                return new IntValue(runner.runParallel(bound));
            }
        }
        public RWSet rwSet() {
            return _bound.rwSet().union(_body.rwSet().completeIfContains(_var));
        }
    }

    /** Serial for */
    public static class SFor implements Exp {
        String _var;
        public Exp _bound, _body;
        public SFor(String var, Exp bound, Exp body) {
            _var = var.intern();
            _bound = bound;
            _body = body;
        }
        public Value eval(Env e, FEnv f) {
            int bound = _bound.eval(e, f).intValue();
            for (int i = 0; i < bound; ++i) {
                _body.eval(ext(e, _var, new IntValue(i)), f);
            }
            return new IntValue(0);
        }
        public RWSet rwSet() {
            return _bound.rwSet().union(_body.rwSet().completeIfContains(_var));
        }
    }


    // helpers arrays-related operations
    public static Value[] allocArray(int size, Value elem) {
        Value [] arr = new Value [size];
        for (int i = 0; i < arr.length; ++i) { arr[i] = elem; }
        return arr;
    }

    public abstract static class LoopRunner {
        public abstract int run(int lower, int upper);
        public int runSerial(final int bound) { return run(0, bound); }
        public int runParallel(final int bound) {
            final int numThreadsInThisLoop = Math.max(1,Math.min(pool().getMaximumPoolSize()-pool().getActiveCount(), bound));
            Collection<Callable<Integer>> tasks =
                new ArrayList<Callable<Integer>>(numThreadsInThisLoop);
            final int chunk_size = bound / numThreadsInThisLoop;
            final int additional_item = bound % numThreadsInThisLoop; // how many threads get one additional item
//            util.IO.addOutput("LoopRunner.runParallel(bound="+bound+"), numThreadsInThisLoop="+numThreadsInThisLoop+
//                              ", chunk_size="+chunk_size+", additional_item = "+additional_item+
//                              ": "+(additional_item)+"x"+(chunk_size+1)+" + "+(numThreadsInThisLoop-additional_item)+"x"+chunk_size);
            for (int i = 0; i < numThreadsInThisLoop; ++i) {
                final int ii = i;
                final LoopRunner tthis = this;
                final int from = (i < additional_item) 
                    ? i*(chunk_size+1) 
                    : additional_item*(chunk_size+1) + (i-additional_item)*chunk_size;
                final int to = (i < additional_item) 
                    ? (i+1)*(chunk_size+1)
                    : additional_item*(chunk_size+1) + (i-additional_item+1)*chunk_size;
                // util.IO.addOutput("Thread "+i+" --> from="+from+", to="+to);
                tasks.add(new Callable<Integer>() { public Integer call() { return tthis.run(from, to); }});
            }
            try {
                List<Future<Integer>> futures = pool().invokeAll(tasks);
                for(Future<Integer> f: futures) {
                    f.get(); // this propagates exceptions out of the child threads
                }
            }
            catch(ExecutionException ee) { throw new EvaluationException(ee); }
            catch(InterruptedException ie) { throw new EvaluationException(ie); }
            return 0;
        }
    }

    /** Random number 0 <= x < 1 */
    public static class Rand implements Exp {
        public Rand() { }
        public Value eval(Env e, FEnv f) {
            return new DoubleValue(getRandom());
        }
        public RWSet rwSet() { return new RWSet(); }
    }
    
    public static final HashMap<Thread,Random> RANDOM_FOR_THREAD = new HashMap<Thread,Random>();
    
    public static double getRandom() {
        Random r = RANDOM_FOR_THREAD.get(Thread.currentThread());
        if (r == null) {
            RANDOM_FOR_THREAD.put(Thread.currentThread(), r = new Random()); // seed?
        }
        return r.nextDouble();
    }
    
    /** Display an ArrayValue of DoubleValues as square grayscale image. */
    public static class DisplayImage implements Exp {
        Exp _image;
        String _title;
        public DisplayImage(Exp image, String title) { _image = image; _title = title; }
        public Value eval(Env e, FEnv f) {
            Value[] a = _image.eval(e, f).arrayValue();
            if (a.length==0) { throw new EvaluationException("Empty image"); }
            int side = (int)Math.sqrt(a.length);
            if (side*side != a.length) {
                throw new EvaluationException("Not a square image: image.length="+a.length+", side="+side);
            }
            double[] image = new double[a.length];
            for(int i=0; i<a.length; ++i) {
                image[i] = Math.max(0, Math.min(1, a[i].doubleValue())); // clamp to 0<=temp<=1                
            }
            util.Misc.displayImages(side, _title, image);
            return new IntValue(0);
        }
        public RWSet rwSet() { return new RWSet(); }
    }
    
    /** Load an image into a ArrayValue of DoubleValues. */
    public static class LoadImage implements Exp {
        String _fileName;
        public LoadImage(String fileName) { _fileName = fileName; }
        public Value eval(Env e, FEnv f) {
            try {
                java.awt.image.BufferedImage image = javax.imageio.ImageIO.read(new java.io.File(_fileName));
                int side = Math.min(image.getWidth(),image.getHeight());
                Value[] a = new Value[side*side];
                java.awt.image.Raster r = image.getData(new java.awt.Rectangle(0,0,side,side));
                int i=0;
                for(int y=0; y<side; ++y) {
                    for(int x=0; x<side; ++x) {
                        double acc = 0;
                        for(int band=0; band<r.getNumBands(); ++band) {
                            acc += r.getSampleDouble(x, y, band) / Math.pow(2, r.getSampleModel().getSampleSize(band));
                        }
                        a[i++] = new DoubleValue(acc/r.getNumBands());
                    }
                }
                return new ArrayValue(a);
            }
            catch(java.io.IOException ioe) { throw new EvaluationException(ioe); }
        }
        public RWSet rwSet() { return new RWSet(); }
    }

    /** Make a Gaussian blur kernel with a given size. */
    public static class GaussianKernel implements Exp {
        Exp _size;
        public GaussianKernel(Exp size) { _size = size; }
        public Value eval(Env e, FEnv f) {
            int size = _size.eval(e, f).intValue();
            double[] kernel = util.Misc.makeKernel(size);
            Value[] a = new Value[kernel.length];
            for(int i=0; i<kernel.length; ++i) {
                a[i] = new DoubleValue(kernel[i]);
            }
            return new ArrayValue(a);
        }
        public RWSet rwSet() { return new RWSet(); }
    }


    public static class EvaluationException extends RuntimeException {
        public EvaluationException() { super(); }
        public EvaluationException(String s) { super(s); }
        public EvaluationException(Throwable cause) { super(cause); }
        public EvaluationException(String s, Throwable cause) { super(s, cause); }
    }
    
    public static class EnvironmentException extends RuntimeException {
        public EnvironmentException() { super(); }
        public EnvironmentException(String s) { super(s); }
    }

    // interface to represent Value -> Value functions
    public static interface ValueFun {
        public Value apply(Value param);
    }
    
    // environment; implemented as a function object from String to Value
    public static interface Env {
        public Value get(String y);
    }
    public static final Env env0 = new Env() {
        public Value get(String s) { throw new EnvironmentException(s); }
    };
    
    public static Env ext(final Env env, final String x, final Value v) {
        return new Env() {
            public Value get(String y) {
                if (x.equals(y)) return v; else return env.get(y);
            }
        };
    }
    
    // function environment; implemented as a function object from String to ValueFun
    public static interface FEnv {
        public ValueFun get(String s);
    }
    public static final FEnv fenv0 = new FEnv() {
        public ValueFun get(String s) { throw new EnvironmentException(); }
    };
    
    public static FEnv fext(final FEnv fenv, final String x, final ValueFun v) {
        return new FEnv() {
            public ValueFun get(String y) {
                if (x.equals(y)) return v; else return fenv.get(y);
            }
        };
    }

    // for computing if a read and write set overlap
    public static boolean rwOverlaps(Env e, RWSet rwSet) {
        if (rwSet.overlaps()) return true;
        for (String rVar : rwSet.readElems()) {
            Value r = e.get(rVar);
            for (String wVar : rwSet.writeElems()) {
                Value w = e.get(wVar);
                if (r instanceof ArrayValue && w instanceof ArrayValue
                    && r.arrayValue() == w.arrayValue())
                    return true;
            }
        }
        return false;
    }

    public static class Declaration {
        public String _fun, _param;
        public Exp _body;
        public Declaration(String fun, String param, Exp body) {
            _fun = fun;
            _param = param;
            _body = body;
        }
        public String fun() { return _fun; }
        public String param() { return _param; }
        public Exp body() { return _body; }
    }
    
    public static class Program {
        public Declaration[] _defs;
        public Exp _body;
        public Program(Exp body, Declaration ... defs) {
            _defs = defs;
            _body = body;
        }
        public Declaration[] defs(){ return _defs; }
        public Exp body() { return _body; }
        
        public Value peval(Env env, FEnv fenv) {
            return peval(env,fenv,0);
        }
        
        private Value peval(final Env env, final FEnv fenv, int defIndex) {
            // match p with
            if (_defs.length<=defIndex) {
                //     Program ([],e) -> eval e env fenv
                return _body.eval(env,fenv);
            }
            else {
                //    |Program (Declaration (s1,s2,e1)::tl,e) ->
                //        let rec f x = eval e1 (ext env s2 x) (ext fenv s1 f)
                //        in peval (Program(tl,e)) env (ext fenv s1 f)
                final Declaration d = _defs[defIndex];
                final String dParam = d._param;
                final String dFun = d._fun;
                final Exp dBody = d._body;
                ValueFun f = new ValueFun() {
                    public Value apply(Value x) {
                        return dBody.eval(ext(env, dParam, x), fext(fenv, dFun, this));
                    }
                };
                return peval(env,fext(fenv, dFun, f),defIndex+1);
            }
        }
    }
    

    // parser
    public static final parser.LintParser<Exp> PARSER = new parser.LintParser<Exp>(Exp.class)
        .addSubParser(new parser.LintParser<Value>(Value.class));
    

    // example terms

    public static Program parsedTermFact = new Program
        (PARSER.parse("(App f (Val (IntValue 10)))"),
         new Declaration
             ("f", "n", PARSER.parse("(If (Equals (Var n) (Val (IntValue 0)))"+
                                     "    (Val (IntValue 1))"+
                                     "    (Mul (Var n)"+
                                     "         (App f (Sub (Var n)"+
                                     "                     (Val (IntValue 1))))))")));
    
    
    public static Program parsedTermFib = new Program
        (PARSER.parse("(App f (Val (IntValue 10)))"),
         new Declaration
             ("f", "n", PARSER.parse("(If (Equals (Var n) (Val (IntValue 0)))"+
                                     "    (Val (IntValue 0))"+
                                     "    (If (Equals (Sub (Var n) (Val (IntValue 1)))"+
                                     "                (Val (IntValue 0)))"+
                                     "        (Val (IntValue 1))"+
                                     "        (Add (App f (Sub (Var n) (Val (IntValue 1))))"+
                                     "             (App f (Sub (Var n) (Val (IntValue 2)))))))")));
    
    /*
     let rec f x = if x=0 then 1 else x*(f(x-1)) in f 10
     let rec f x = if x=0 then 1 else x*(f(x-1)) in f (10/2)
     */
    
    public static void main(String[] args) {
        Value i = parsedTermFact.peval(env0, fenv0);
        System.out.println("fact(10) = "+i);
        
        i = parsedTermFib.peval(env0, fenv0);
        System.out.println("fib(10) = "+i);
    }
}
