package gadts;

import java.util.*;

// read and write sets of arrays
public class RWSet {
    public static class NoSuchElementException extends RuntimeException { }
    public static class UnsupportedOperationException extends RuntimeException { }
    
    public abstract class StringSet implements Iterable<String> /* Set<String> */ {
        public separable abstract boolean isComplete ();
        public separable abstract boolean contains (String s);
        public separable abstract StringSet unionOne (String s);
        public separable abstract StringSet remove (String s);
        public separable abstract StringSet union (StringSet set);
        public separable abstract boolean overlaps (StringSet set);
        public separable abstract int size ();
        public separable Iterator<String> iterator () { return new StringSetIterator (this); }
        public separable String toString() {
            if (size()==0) return "{}";
            String acc = "{";
            for(String s: this) { acc += " "+s; }
            return acc+" }";
        }
        
//        public boolean add(String e) { throw new UnsupportedOperationException(); }
//        public boolean addAll(Collection<? extends String> c) { throw new UnsupportedOperationException(); }
//        public boolean remove(Object o) { throw new UnsupportedOperationException(); }
//        public boolean removeAll(Collection<?> c) { throw new UnsupportedOperationException(); }
//        public boolean retainAll(Collection<?> c) { throw new UnsupportedOperationException(); }
//        public void clear() { throw new UnsupportedOperationException(); }
    }

    public class ElemStringSet extends StringSet {
        separable ElemStringSet (String elem, StringSet next) {
            _elem = elem; // elem needs to have been interned, so we can use == below
            _next = next;
        }
        String _elem;
        StringSet _next;

        public separable boolean isComplete () { return false; }
        public separable int size () { return _next.size() + 1; }
        public separable boolean contains (String s) {
            return s == _elem || (_next.contains (s));
        }
        public separable StringSet unionOne (String s) {
            if (contains (s)) return this;
            else return new ElemStringSet (s, this);
        }
        public separable StringSet remove (String s) {
            if (s == _elem) return _next;
            else return new ElemStringSet (_elem, _next.remove (s));
        }
        public separable StringSet union (StringSet set) {
            if (set instanceof EmptyStringSet) return this;
            else if (set instanceof CompleteStringSet) return set;
            else return unionOne(((ElemStringSet)set)._elem).union(((ElemStringSet)set)._next);
        }
        public separable boolean overlaps (StringSet set) {
            return set.contains (_elem) || _next.overlaps (set);
        }
    }

    public class CompleteStringSet extends StringSet {
        public separable boolean isComplete () { return true; }
        public separable int size () { return 0; }
        public separable boolean contains (String s) { return true; }
        public separable StringSet unionOne (String s) { return this; }
        public separable StringSet remove (String s) { return this; }
        public separable StringSet union (StringSet set) { return this; }
        public separable boolean overlaps (StringSet set) {
            if (set instanceof EmptyStringSet) return false;
            else return true;
        }
    }

    public class EmptyStringSet extends StringSet {
        public separable boolean isComplete () { return false; }
        public separable int size () { return 0; }
        public separable boolean contains (String s) { return false; }
        public separable StringSet unionOne (String s) { return new ElemStringSet (s, this); }
        public separable StringSet remove (String s) { return this; }
        public separable StringSet union (StringSet set) { return set; }
        public separable boolean overlaps (StringSet set) { return false; }
    }

    public class StringSetIterator implements Iterator<String> {
        String[] strings;
        int cur_index;

        public separable StringSetIterator (StringSet set) {
            strings = new String [set.size ()];
            setStrings (set, 0);
            cur_index = 0;
        }

        separable void setStrings (StringSet set, int i) {
            if (set instanceof ElemStringSet) {
                strings [i] = ((ElemStringSet)set)._elem;
                setStrings (((ElemStringSet)set)._next, i+1);
            }
        }

        public separable boolean hasNext () {
            return cur_index < strings.length;
        }

        public separable String next () {
            if (hasNext ()) {
                return strings[cur_index++];
            } else {
                throw new NoSuchElementException ();
            }
        }

        public separable void remove () {
            throw new UnsupportedOperationException ();
        }
    }


    public separable RWSet (boolean is_complete) {
        StringSet s = is_complete ? new CompleteStringSet () : new EmptyStringSet ();
        read = s;
        write = s;
    }
    public separable RWSet () {
        read = new EmptyStringSet ();
        write = new EmptyStringSet ();
    }
    public separable RWSet (String var, boolean is_read) {
        read = new EmptyStringSet ();
        write = new EmptyStringSet ();
        if (is_read) {
            read = read.unionOne (var);
        } else {
            write = write.unionOne (var);
        }
    }
    protected separable RWSet (StringSet r, StringSet w) {
        read = r;
        write = w;
    }

    StringSet read;
    StringSet write;

    public separable boolean containsR (String s) {
        return read.contains (s);
    }
    public separable boolean containsW (String s) {
        return write.contains (s);
    }
    public separable RWSet addR (String s) {
        return new RWSet (read.unionOne (s), write);
    }
    public separable RWSet addW (String s) {
        return new RWSet (read, write.unionOne (s));
    }
    public separable RWSet completeR () {
        return new RWSet (new CompleteStringSet (), write);
    }
    public separable RWSet completeW () {
        return new RWSet (read, new CompleteStringSet ());
    }
    public separable RWSet removeR (String s) {
        return new RWSet (read.remove (s), write);
    }
    public separable RWSet removeW (String s) {
        return new RWSet (read, write.remove (s));
    }
    public separable RWSet union (RWSet set) {
        return new RWSet (read.union (set.read), write.union (set.write));
    }
    public separable RWSet completeIfContains (String s) {
        StringSet new_read = read.contains (s) ? new CompleteStringSet () : read;
        StringSet new_write = write.contains (s) ? new CompleteStringSet () : write;
        return new RWSet (new_read, new_write);
    }
    public separable boolean overlaps () { return read.overlaps (write); }
    public separable StringSet readElems () { return read; }
    public separable StringSet writeElems () { return write; }
    
    public String toString() {
        return "(RWSet read="+read+" write="+write+")";
    }
}
