package gadts;

import edu.rice.cs.mint.runtime.Code;
import java.util.concurrent.*;
import java.util.*;

public class StagedDataTypes {
    public static int NUM_THREADS = Runtime.getRuntime().availableProcessors();
    public static volatile ThreadPoolExecutor pool = null;
    public static synchronized ThreadPoolExecutor pool() {
        if (pool==null) {
            pool = new ThreadPoolExecutor(NUM_THREADS, // number of core threads
                                          NUM_THREADS, // maximum number
                                          0, TimeUnit.SECONDS, // keep alive > core
                                          new LinkedBlockingQueue<Runnable>());
        }
        return pool;
    }

    public static class WrongTypeException extends RuntimeException {
        public separable WrongTypeException() { super(); }
        public separable WrongTypeException(String msg) { super(msg); }
        public static separable <T> T returningThrow() { throw new WrongTypeException(); }
    }
    public static class EnvironmentException extends RuntimeException {
        separable EnvironmentException (String var) { super(var); }
        public static separable <T> T returningThrow(String var) { throw new EnvironmentException(var); }
    }
    public static class EvaluationException extends RuntimeException {
        public EvaluationException() { super(); }
        public EvaluationException(String s) { super(s); }
        public EvaluationException(Throwable cause) { super(cause); }
        public EvaluationException(String s, Throwable cause) { super(s, cause); }
    }

    public static class Value {
        public separable int intValue() { throw new WrongTypeException(); }
        public separable double doubleValue() { throw new WrongTypeException(); }
        public separable boolean booleanValue() { throw new WrongTypeException(); }
        public separable Value[] arrayValue() { throw new WrongTypeException(); }
        public boolean compare (Value other) {
            if (this instanceof IntValue && other instanceof IntValue) {
                return intValue() == other.intValue ();
            }
            else if (this instanceof DoubleValue && other instanceof DoubleValue) {
                return doubleValue() == other.doubleValue ();
            }
            else if (this instanceof BooleanValue && other instanceof BooleanValue) {
                return booleanValue() == other.booleanValue ();
            }
            else if (this instanceof ArrayValue && other instanceof ArrayValue) {
                return arrayValue() == other.arrayValue ();
            }
            else {
                return false;
            }

        }
    }

    public static class IntValue extends Value {
        private int _data;
        public IntValue(int data) { _data = data; }
        public separable int intValue() { return _data; }
        public separable double doubleValue() { return _data; }
        public String toString() { return _data+":IntValue"; }
    }

    public static class DoubleValue extends Value {
        private double _data;
        public DoubleValue(double data) { _data = data; }
        public separable double doubleValue() { return _data; }
        public String toString() { return _data+":DoubleValue"; }
    }
    
    public static class BooleanValue extends Value {
        private boolean _data;
        public BooleanValue(boolean data) { _data = data; }
        public separable boolean booleanValue() { return _data; }
        public String toString() { return _data+":BooleanValue"; }
    }
    
    public static class ArrayValue extends Value {
        private Value[] _data;
        public ArrayValue(Value[] data) { _data = data; }
        public separable Value[] arrayValue() { return _data; }
        public String toString() { return java.util.Arrays.toString(_data)+":ArrayValue"; }
    }

    /**
     * Lattice elements for our type lattice
     */

    // a polymorphic code object; generally code of an exception
    public static interface AnyCode {
        public separable <X> Code<X> cast (Class<X> c);
    }

    public static separable Code<Value> intCodeToValue (Code<Integer> c) {
        return <| (Value) new IntValue (`c) |>;
    }
    public static separable Code<Value> booleanCodeToValue (Code<Boolean> c) {
        return <| (Value) new BooleanValue (`c) |>;
    }

    public static separable Code<Value> arrayCodeToValue (Code<Value[]> c) {
        return <| (Value) new ArrayValue (`c) |>;
    }

    public static separable Code<Value> doubleCodeToValue (Code<Double> c) {
        return <| (Value) new DoubleValue (`c) |>;
    }

    // interfaces for functions and separable functions
    public static interface Fun<T,U> {
        public U apply (T input);
    }
    public static interface SepFun<T,U> extends Fun<T,U> {
        public separable U apply (T input);
    }

    /**
     * helper classes for DSLType methods
     */

    // a T (which is generally Code<?> or AnyCode) along with a DSLType<T>
    public static class SValue<T> {
        DSLType<T> tp;
        T code;
        separable SValue (DSLType<T> tp, T code) {
            this.tp = tp; this.code = code;
        }
        public static SValue<Code<Integer>> make (final int i) {
            return new SValue<Code<Integer>> (new DSLTypeInteger(), <| i |>);
        }
        public static SValue<Code<Boolean>> make (final boolean b) {
            return new SValue<Code<Boolean>> (new DSLTypeBoolean(), <| b |>);
        }
        public static SValue<Code<Double>> make (final double d) {
            return new SValue<Code<Double>> (new DSLTypeDouble(), <| d |>);
        }
    }

    // helper class for the visitor pattern (i.e. matching) on DSLTypes
    public static interface TypeVisitor<R> {
        public R visitTop (IR<AnyCode> ir);
        public R visitInteger (IR<Code<Integer>> ir);
        public R visitBoolean (IR<Code<Boolean>> ir);
        public R visitArray (IR<Code<Value[]>> ir);
        public R visitDouble (IR<Code<Double>> ir);
        public R visitBottom (IR<Code<Value>> ir);
    }
    public static interface TypeVisitor2<R> {
        public R visitTop (IR<AnyCode> ir1, IR<AnyCode> ir2);
        public R visitInteger (IR<Code<Integer>> ir1, IR<Code<Integer>> ir2);
        public R visitBoolean (IR<Code<Boolean>> ir1, IR<Code<Boolean>> ir2);
        public R visitArray (IR<Code<Value[]>> ir1, IR<Code<Value[]>> ir2);
        public R visitDouble (IR<Code<Double>> ir1, IR<Code<Double>> ir2);
        public R visitBottom (IR<Code<Value>> ir1, IR<Code<Value>> ir2);
    }
    public static interface TypeVisitorCode<R> {
        public separable R visitTop (AnyCode c);
        public separable R visitInteger (Code<Integer> c);
        public separable R visitBoolean (Code<Boolean> c);
        public separable R visitArray (Code<Value[]> c);
        public separable R visitDouble (Code<Double> c);
        public separable R visitBottom (Code<Value> c);
    }


    // this represents the fact that both T1 and T2 are subtypes of U
    public static class JoinReturn<T> {
        public DSLType<T> tp;
        public IR<T> ir1;
        public IR<T> ir2;
        public JoinReturn (DSLType<T> tp, IR<T> ir1, IR<T> ir2) {
            this.tp = tp; this.ir1 = ir1; this.ir2 = ir2;
        }
    }


    /**
     * The definitions of the DSL types
     * DSLType<T> is a DSL type that can produce a Code<T>
     */

    public static abstract class DSLType<T> {
        // convert types to string in separable contexts
        public separable abstract String toString ();

        // pattern match on the type of an IR using a TypeVisitor
        public abstract <R> R accept (IR<T> ir, TypeVisitor<R> v);
        public abstract <R> R accept2 (IR<T> ir1, IR<T> ir2, TypeVisitor2<R> v);
        public separable abstract <R> R acceptCode (T c,
                                                    TypeVisitorCode<R> v);

        // try to cast ir to type U, returning null on failure
        public abstract <U> IR<T> tryCast (IR<U> ir, boolean unboxOnFailure);

        // try to cast code c of type U to type T, returning null on failure
        public separable abstract <U> T tryCastCode (DSLType<U> tp, U c);
        public separable <U> T tryCastCode (SValue<U> svalue) {
            return tryCastCode (svalue.tp, svalue.code);
        }

        // find a DSLType<?> that both ir1 and ir2 can be downcast to
        public abstract <U> JoinReturn<?> join (IR<T> ir1, IR<U> c2);

        // create code that compares values of type T
        public separable abstract Code<Boolean> valueCmp (T c1, T c2);
    }


    // Top type, i.e., polymorphic AnyCode
    public static class DSLTypeTop extends DSLType<AnyCode> {
        public separable String toString () { return "Top"; }
        public <R> R accept (IR<AnyCode> ir, TypeVisitor<R> v) {
            return v.visitTop (ir);
        }
        public separable <R> R acceptCode (AnyCode c, TypeVisitorCode<R> v) {
            return v.visitTop (c);
        }
        public <R> R accept2 (IR<AnyCode> ir1, IR<AnyCode> ir2, TypeVisitor2<R> v) {
            return v.visitTop (ir1, ir2);
        }
        public <U> IR<AnyCode> tryCast (final IR<U> ir,
                                        final boolean unboxOnFailure) {
            return ir.tp.accept (ir, new TypeVisitor<IR<AnyCode>>() {
                    public IR<AnyCode> visitTop (IR<AnyCode> ir) {
                        return ir;
                    }
                    public IR<AnyCode> visitInteger (IR<Code<Integer>> ir) {
                        return unboxOnFailure ?
                            new ErrorIRAny<Integer>(ir, Integer.class) : null;
                    }
                    public IR<AnyCode> visitBoolean (IR<Code<Boolean>> ir) {
                        return unboxOnFailure ?
                            new ErrorIRAny<Boolean>(ir, Boolean.class) : null;
                    }
                    public IR<AnyCode> visitArray (IR<Code<Value[]>> ir) {
                        return unboxOnFailure ?
                            new ErrorIRAny<Value[]>(ir, Value[].class) : null;
                    }
                    public IR<AnyCode> visitDouble (IR<Code<Double>> ir) {
                        return unboxOnFailure ?
                            new ErrorIRAny<Double>(ir, Double.class) : null;
                    }
                    public IR<AnyCode> visitBottom (IR<Code<Value>> ir) {
                        return unboxOnFailure ?
                            new ErrorIRAny<Value>(ir, Value.class) : null;
                    }
                });
        }
        public separable <U> AnyCode tryCastCode (DSLType<U> tp, final U c) {
            return tp.acceptCode (c, new TypeVisitorCode<AnyCode>() {
                    public separable AnyCode visitTop (AnyCode c) {
                        return c;
                    }
                    public separable AnyCode visitInteger (Code<Integer> c) {
                        return null;
                    }
                    public separable AnyCode visitBoolean (Code<Boolean> c) {
                        return null;
                    }
                    public separable AnyCode visitArray (Code<Value[]> c) {
                        return null;
                    }
                    public separable AnyCode visitDouble (Code<Double> c) {
                        return null;
                    }
                    public separable AnyCode visitBottom (Code<Value> c) {
                        return null;
                    }
                });
        }
        public <U> JoinReturn<?> join (final IR<AnyCode> ir1, IR<U> ir2) {
            return ir2.tp.accept (ir2, new TypeVisitor<JoinReturn<?>>() {
                    public JoinReturn<?> visitTop (IR<AnyCode> ir) {
                        return new JoinReturn<AnyCode>
                            (new DSLTypeTop (), ir1, ir);
                    }
                    public JoinReturn<?> visitInteger (IR<Code<Integer>> ir) {
                        return new JoinReturn<Code<Integer>>
                            (new DSLTypeInteger (), new BoxAnyInt (ir1), ir);
                    }
                    public JoinReturn<?> visitBoolean (IR<Code<Boolean>> ir) {
                        return new JoinReturn<Code<Boolean>>
                            (new DSLTypeBoolean (), new BoxAnyBool (ir1), ir);
                    }
                    public JoinReturn<?> visitArray (IR<Code<Value[]>> ir) {
                        return new JoinReturn<Code<Value[]>>
                            (new DSLTypeArray (), new BoxAnyArray (ir1), ir);
                    }
                    public JoinReturn<?> visitDouble (IR<Code<Double>> ir) {
                        return new JoinReturn<Code<Double>>
                            (new DSLTypeDouble (), new BoxAnyDouble (ir1), ir);
                    }
                    public JoinReturn<?> visitBottom (IR<Code<Value>> ir) {
                        return new JoinReturn<Code<Value>>
                            (new DSLTypeBottom (), new BoxAnyVal (ir1), ir);
                    }
                });
        }
        public separable Code<Boolean> valueCmp (AnyCode c1, AnyCode c2) {
            return <| `(c1.cast (Integer.class)) == `(c2.cast (Integer.class)) |>; // MGR: why Integer?
        }
    }


    // Integer type
    public static class DSLTypeInteger extends DSLType<Code<Integer>> {
        public separable String toString () { return "Int"; }
        public <R> R accept (IR<Code<Integer>> ir, TypeVisitor<R> v) {
            return v.visitInteger (ir);
        }
        public separable <R> R acceptCode (Code<Integer> c, TypeVisitorCode<R> v) {
            return v.visitInteger (c);
        }
        public <R> R accept2 (IR<Code<Integer>> ir1, IR<Code<Integer>> ir2, TypeVisitor2<R> v) {
            return v.visitInteger (ir1, ir2);
        }
        public <U> IR<Code<Integer>> tryCast (final IR<U> ir,
                                              final boolean unboxOnFailure) {
            return ir.tp.accept (ir, new TypeVisitor<IR<Code<Integer>>>() {
                    public IR<Code<Integer>> visitTop (IR<AnyCode> ir) {
                        return new BoxAnyInt (ir);
                    }
                    public IR<Code<Integer>> visitInteger (IR<Code<Integer>> ir) {
                        return ir;
                    }
                    public IR<Code<Integer>> visitBoolean (IR<Code<Boolean>> ir) {
                        return unboxOnFailure ?
                            new ErrorIRCode<Integer,Boolean>
                            (new DSLTypeInteger (), Integer.class, ir) : null;
                    }
                    public IR<Code<Integer>> visitArray (IR<Code<Value[]>> ir) {
                        return unboxOnFailure ?
                            new ErrorIRCode<Integer,Value[]>
                            (new DSLTypeInteger (), Integer.class, ir) : null;
                    }
                    public IR<Code<Integer>> visitDouble (IR<Code<Double>> ir) {
                        return unboxOnFailure ?
                            new ErrorIRCode<Integer,Double>
                            (new DSLTypeInteger (), Integer.class, ir) : null;
                    }
                    public IR<Code<Integer>> visitBottom (IR<Code<Value>> ir) {
                        return unboxOnFailure ? new UnboxIntIR (ir) : null;
                    }
                });
        }
        public separable <U> Code<Integer> tryCastCode (DSLType<U> tp, final U c) {
            return tp.acceptCode (c, new TypeVisitorCode<Code<Integer>>() {
                    public separable Code<Integer> visitTop (AnyCode c) {
                        return c.cast (Integer.class);
                    }
                    public separable Code<Integer> visitInteger (Code<Integer> c) {
                        return c;
                    }
                    public separable Code<Integer> visitBoolean (Code<Boolean> c) {
                        return null;
                    }
                    public separable Code<Integer> visitArray (Code<Value[]> c) {
                        return null;
                    }
                    public separable Code<Integer> visitDouble (Code<Double> c) {
                        return null;
                    }
                    public separable Code<Integer> visitBottom (Code<Value> c) {
                        return null;
                    }
                });
        }
        public <U> JoinReturn<?> join (final IR<Code<Integer>> ir1, IR<U> ir2) {
            return ir2.tp.accept (ir2, new TypeVisitor<JoinReturn<?>>() {
                    public JoinReturn<?> visitTop (IR<AnyCode> ir) {
                        return new JoinReturn<Code<Integer>>
                            (new DSLTypeInteger (),
                             ir1, new BoxAnyInt (ir));
                    }
                    public JoinReturn<?> visitInteger (IR<Code<Integer>> ir) {
                        return new JoinReturn<Code<Integer>>
                            (new DSLTypeInteger (), ir1, ir);
                    }
                    public JoinReturn<?> visitBoolean (IR<Code<Boolean>> ir) {
                        return new JoinReturn<Code<Value>>
                            (new DSLTypeBottom (),
                             new BoxIntValue (ir1), new BoxBoolValue (ir));
                    }
                    public JoinReturn<?> visitArray (IR<Code<Value[]>> ir) {
                        return new JoinReturn<Code<Value>>
                            (new DSLTypeBottom (),
                             new BoxIntValue (ir1), new BoxArrayValue (ir));
                    }
                    public JoinReturn<?> visitDouble (IR<Code<Double>> ir) {
                        return new JoinReturn<Code<Double>>
                            (new DSLTypeDouble (),
                             new BoxIntDouble (ir1), ir);
                    }
                    public JoinReturn<?> visitBottom (IR<Code<Value>> ir) {
                        return new JoinReturn<Code<Value>>
                            (new DSLTypeBottom (), new BoxIntValue (ir1), ir);
                    }
                });
        }
        public separable Code<Boolean> valueCmp (Code<Integer> c1,
                                                 Code<Integer> c2) {
            return <| `c1 == `c2 |>;
        }
    }


    // Boolean type
    public static class DSLTypeBoolean extends DSLType<Code<Boolean>> {
        public separable String toString () { return "Bool"; }
        public <R> R accept (IR<Code<Boolean>> ir, TypeVisitor<R> v) {
            return v.visitBoolean (ir);
        }
        public separable <R> R acceptCode (Code<Boolean> c, TypeVisitorCode<R> v) {
            return v.visitBoolean (c);
        }
        public <R> R accept2 (IR<Code<Boolean>> ir1, IR<Code<Boolean>> ir2, TypeVisitor2<R> v) {
            return v.visitBoolean (ir1, ir2);
        }
        public <U> IR<Code<Boolean>> tryCast (final IR<U> ir,
                                              final boolean unboxOnFailure) {
            return ir.tp.accept (ir, new TypeVisitor<IR<Code<Boolean>>>() {
                    public IR<Code<Boolean>> visitTop (IR<AnyCode> ir) {
                        return new BoxAnyBool (ir);
                    }
                    public IR<Code<Boolean>> visitInteger (IR<Code<Integer>> ir) {
                        return unboxOnFailure ?
                            new ErrorIRCode<Boolean,Integer>
                            (new DSLTypeBoolean (), Boolean.class, ir) : null;
                    }
                    public IR<Code<Boolean>> visitBoolean (IR<Code<Boolean>> ir) {
                        return ir;
                    }
                    public IR<Code<Boolean>> visitArray (IR<Code<Value[]>> ir) {
                        return unboxOnFailure ?
                            new ErrorIRCode<Boolean,Value[]>
                            (new DSLTypeBoolean (), Boolean.class, ir) : null;
                    }
                    public IR<Code<Boolean>> visitDouble (IR<Code<Double>> ir) {
                        return unboxOnFailure ?
                            new ErrorIRCode<Boolean,Double>
                            (new DSLTypeBoolean (), Boolean.class, ir) : null;
                    }
                    public IR<Code<Boolean>> visitBottom (IR<Code<Value>> ir) {
                        return unboxOnFailure ? new UnboxBoolIR(ir) : null;
                    }
                });
        }
        public separable <U> Code<Boolean> tryCastCode (DSLType<U> tp, final U c) {
            return tp.acceptCode (c, new TypeVisitorCode<Code<Boolean>>() {
                    public separable Code<Boolean> visitTop (AnyCode c) {
                        return c.cast (Boolean.class);
                    }
                    public separable Code<Boolean> visitInteger (Code<Integer> c) {
                        return null;
                    }
                    public separable Code<Boolean> visitBoolean (Code<Boolean> c) {
                        return c;
                    }
                    public separable Code<Boolean> visitArray (Code<Value[]> c) {
                        return null;
                    }
                    public separable Code<Boolean> visitDouble (Code<Double> c) {
                        return null;
                    }
                    public separable Code<Boolean> visitBottom (Code<Value> c) {
                        return null;
                    }
                });
        }
        public <U> JoinReturn<?> join (final IR<Code<Boolean>> ir1, IR<U> ir2) {
            return ir2.tp.accept (ir2, new TypeVisitor<JoinReturn<?>>() {
                    public JoinReturn<?> visitTop (IR<AnyCode> ir) {
                        return new JoinReturn<Code<Boolean>>
                            (new DSLTypeBoolean (),
                             ir1, new BoxAnyBool (ir));
                    }
                    public JoinReturn<?> visitInteger (IR<Code<Integer>> ir) {
                        return new JoinReturn<Code<Value>>
                            (new DSLTypeBottom (),
                             new BoxBoolValue (ir1), new BoxIntValue (ir));
                    }
                    public JoinReturn<?> visitBoolean (IR<Code<Boolean>> ir) {
                        return new JoinReturn<Code<Boolean>>
                            (new DSLTypeBoolean (), ir1, ir);
                    }
                    public JoinReturn<?> visitArray (IR<Code<Value[]>> ir) {
                        return new JoinReturn<Code<Value>>
                            (new DSLTypeBottom (),
                             new BoxBoolValue (ir1), new BoxArrayValue (ir));
                    }
                    public JoinReturn<?> visitDouble (IR<Code<Double>> ir) {
                        return new JoinReturn<Code<Value>>
                            (new DSLTypeBottom (),
                             new BoxBoolValue (ir1), new BoxDoubleValue (ir));
                    }
                    public JoinReturn<?> visitBottom (IR<Code<Value>> ir) {
                        return new JoinReturn<Code<Value>>
                            (new DSLTypeBottom (), new BoxBoolValue (ir1), ir);
                    }
                });
        }
        public separable Code<Boolean> valueCmp (Code<Boolean> c1,
                                                 Code<Boolean> c2) {
            return <| `c1 == `c2 |>;
        }
    }


    // Array type
    public static class DSLTypeArray extends DSLType<Code<Value[]>> {
        public separable String toString () { return "Array"; }
        public <R> R accept (IR<Code<Value[]>> ir, TypeVisitor<R> v) {
            return v.visitArray (ir);
        }
        public separable <R> R acceptCode (Code<Value[]> c, TypeVisitorCode<R> v) {
            return v.visitArray (c);
        }
        public <R> R accept2 (IR<Code<Value[]>> ir1, IR<Code<Value[]>> ir2, TypeVisitor2<R> v) {
            return v.visitArray (ir1, ir2);
        }
        public <U> IR<Code<Value[]>> tryCast (final IR<U> ir,
                                              final boolean unboxOnFailure) {
            return ir.tp.accept (ir, new TypeVisitor<IR<Code<Value[]>>>() {
                    public IR<Code<Value[]>> visitTop (IR<AnyCode> ir) {
                        return new BoxAnyArray (ir);
                    }
                    public IR<Code<Value[]>> visitInteger (IR<Code<Integer>> ir) {
                        return unboxOnFailure ?
                            new ErrorIRCode<Value[],Integer>
                            (new DSLTypeArray (), Value[].class, ir) : null;
                    }
                    public IR<Code<Value[]>> visitBoolean (IR<Code<Boolean>> ir) {
                        return unboxOnFailure ?
                            new ErrorIRCode<Value[],Boolean>
                            (new DSLTypeArray (), Value[].class, ir) : null;
                    }
                    public IR<Code<Value[]>> visitArray (IR<Code<Value[]>> ir) {
                        return ir;
                    }
                    public IR<Code<Value[]>> visitDouble (IR<Code<Double>> ir) {
                        return unboxOnFailure ?
                            new ErrorIRCode<Value[],Double>
                            (new DSLTypeArray (), Value[].class, ir) : null;
                    }
                    public IR<Code<Value[]>> visitBottom (IR<Code<Value>> ir) {
                        return unboxOnFailure ? new UnboxArrayIR(ir) : null;
                    }
                });
        }
        public separable <U> Code<Value[]> tryCastCode (DSLType<U> tp, final U c) {
            return tp.acceptCode (c, new TypeVisitorCode<Code<Value[]>>() {
                    public separable Code<Value[]> visitTop (AnyCode c) {
                        return c.cast (Value[].class);
                    }
                    public separable Code<Value[]> visitInteger (Code<Integer> c) {
                        return null;
                    }
                    public separable Code<Value[]> visitBoolean (Code<Boolean> c) {
                        return null;
                    }
                    public separable Code<Value[]> visitArray (Code<Value[]> c) {
                        return c;
                    }
                    public separable Code<Value[]> visitDouble (Code<Double> c) {
                        return null;
                    }
                    public separable Code<Value[]> visitBottom (Code<Value> c) {
                        return null;
                    }
                });
        }
        public <U> JoinReturn<?> join (final IR<Code<Value[]>> ir1, IR<U> ir2) {
            return ir2.tp.accept (ir2, new TypeVisitor<JoinReturn<?>>() {
                    public JoinReturn<?> visitTop (IR<AnyCode> ir) {
                        return new JoinReturn<Code<Value[]>>
                            (new DSLTypeArray (),
                             ir1, new BoxAnyArray (ir));
                    }
                    public JoinReturn<?> visitInteger (IR<Code<Integer>> ir) {
                        return new JoinReturn<Code<Value>>
                            (new DSLTypeBottom (),
                             new BoxArrayValue (ir1), new BoxIntValue (ir));
                    }
                    public JoinReturn<?> visitBoolean (IR<Code<Boolean>> ir) {
                        return new JoinReturn<Code<Value>>
                            (new DSLTypeBottom (),
                             new BoxArrayValue (ir1), new BoxBoolValue (ir));
                    }
                    public JoinReturn<?> visitArray (IR<Code<Value[]>> ir) {
                        return new JoinReturn<Code<Value[]>>
                            (new DSLTypeArray (), ir1, ir);
                    }
                    public JoinReturn<?> visitDouble (IR<Code<Double>> ir) {
                        return new JoinReturn<Code<Value>>
                            (new DSLTypeBottom (),
                             new BoxArrayValue (ir1), new BoxDoubleValue (ir));
                    }
                    public JoinReturn<?> visitBottom (IR<Code<Value>> ir) {
                        return new JoinReturn<Code<Value>>
                            (new DSLTypeBottom (), new BoxArrayValue (ir1), ir);
                    }
                });
        }
        public separable Code<Boolean> valueCmp (Code<Value[]> c1,
                                                 Code<Value[]> c2) {
            return <| `c1 == `c2 |>;
        }
    }


    // Double type
    public static class DSLTypeDouble extends DSLType<Code<Double>> {
        public separable String toString () { return "Double"; }
        public <R> R accept (IR<Code<Double>> ir, TypeVisitor<R> v) {
            return v.visitDouble (ir);
        }
        public separable <R> R acceptCode (Code<Double> c, TypeVisitorCode<R> v) {
            return v.visitDouble (c);
        }
        public <R> R accept2 (IR<Code<Double>> ir1, IR<Code<Double>> ir2, TypeVisitor2<R> v) {
            return v.visitDouble (ir1, ir2);
        }
        public <U> IR<Code<Double>> tryCast (final IR<U> ir,
                                             final boolean unboxOnFailure) {
            return ir.tp.accept (ir, new TypeVisitor<IR<Code<Double>>>() {
                    public IR<Code<Double>> visitTop (IR<AnyCode> ir) {
                        return new BoxAnyDouble (ir);
                    }
                    public IR<Code<Double>> visitInteger (IR<Code<Integer>> ir) {
                        return new BoxIntDouble (ir);
                    }
                    public IR<Code<Double>> visitBoolean (IR<Code<Boolean>> ir) {
                        return unboxOnFailure ?
                            new ErrorIRCode<Double,Boolean>
                            (new DSLTypeDouble (), Double.class, ir) : null;
                    }
                    public IR<Code<Double>> visitArray (IR<Code<Value[]>> ir) {
                        return unboxOnFailure ?
                            new ErrorIRCode<Double,Value[]>
                            (new DSLTypeDouble (), Double.class, ir) : null;
                    }
                    public IR<Code<Double>> visitDouble (IR<Code<Double>> ir) {
                        return ir;
                    }
                    public IR<Code<Double>> visitBottom (IR<Code<Value>> ir) {
                        return unboxOnFailure ? new UnboxDoubleIR(ir) : null;
                    }
                });
        }
        public separable <U> Code<Double> tryCastCode (DSLType<U> tp, final U c) {
            return tp.acceptCode (c, new TypeVisitorCode<Code<Double>>() {
                    public separable Code<Double> visitTop (AnyCode c) {
                        return c.cast (Double.class);
                    }
                    public separable Code<Double> visitInteger (Code<Integer> c) {
                        return <| (double)(`c) |>;
                    }
                    public separable Code<Double> visitBoolean (Code<Boolean> c) {
                        return null;
                    }
                    public separable Code<Double> visitArray (Code<Value[]> c) {
                        return null;
                    }
                    public separable Code<Double> visitDouble (Code<Double> c) {
                        return c;
                    }
                    public separable Code<Double> visitBottom (Code<Value> c) {
                        return null;
                    }
                });
        }
        public <U> JoinReturn<?> join (final IR<Code<Double>> ir1, IR<U> ir2) {
            return ir2.tp.accept (ir2, new TypeVisitor<JoinReturn<?>>() {
                    public JoinReturn<?> visitTop (IR<AnyCode> ir) {
                        return new JoinReturn<Code<Double>>
                            (new DSLTypeDouble (),
                             ir1, new BoxAnyDouble (ir));
                    }
                    public JoinReturn<?> visitInteger (IR<Code<Integer>> ir) {
                        return new JoinReturn<Code<Value>>
                            (new DSLTypeBottom (),
                             new BoxDoubleValue (ir1), new BoxIntValue (ir));
                    }
                    public JoinReturn<?> visitBoolean (IR<Code<Boolean>> ir) {
                        return new JoinReturn<Code<Value>>
                            (new DSLTypeBottom (),
                             new BoxDoubleValue (ir1), new BoxBoolValue (ir));
                    }
                    public JoinReturn<?> visitArray (IR<Code<Value[]>> ir) {
                        return new JoinReturn<Code<Value>>
                            (new DSLTypeBottom (),
                             new BoxDoubleValue (ir1), new BoxArrayValue (ir));
                    }
                    public JoinReturn<?> visitDouble (IR<Code<Double>> ir) {
                        return new JoinReturn<Code<Double>>
                            (new DSLTypeDouble (), ir1, ir);
                    }
                    public JoinReturn<?> visitBottom (IR<Code<Value>> ir) {
                        return new JoinReturn<Code<Value>>
                            (new DSLTypeBottom (), new BoxDoubleValue (ir1), ir);
                    }
                });
        }
        public separable Code<Boolean> valueCmp (Code<Double> c1,
                                                 Code<Double> c2) {
            return <| `c1 == `c2 |>;
        }
    }

    
    // Bottom type, i.e., Value
    public static class DSLTypeBottom extends DSLType<Code<Value>> {
        public separable String toString () { return "Bottom"; }
        public <R> R accept (IR<Code<Value>> ir, TypeVisitor<R> v) {
            return v.visitBottom (ir);
        }
        public separable <R> R acceptCode (Code<Value> c, TypeVisitorCode<R> v) {
            return v.visitBottom (c);
        }
        public <R> R accept2 (IR<Code<Value>> ir1, IR<Code<Value>> ir2, TypeVisitor2<R> v) {
            return v.visitBottom (ir1, ir2);
        }
        public <U> IR<Code<Value>> tryCast (IR<U> ir,
                                            final boolean unboxOnFailure) {
            return ir.tp.accept (ir, new TypeVisitor<IR<Code<Value>>>() {
                    public IR<Code<Value>> visitTop (IR<AnyCode> ir) {
                        return new BoxAnyVal (ir);
                    }
                    public IR<Code<Value>> visitInteger (IR<Code<Integer>> ir) {
                        return new BoxIntValue (ir);
                    }
                    public IR<Code<Value>> visitBoolean (IR<Code<Boolean>> ir) {
                        return new BoxBoolValue (ir);
                    }
                    public IR<Code<Value>> visitArray (IR<Code<Value[]>> ir) {
                        return new BoxArrayValue (ir);
                    }
                    public IR<Code<Value>> visitDouble (IR<Code<Double>> ir) {
                        return new BoxDoubleValue (ir);
                    }
                    public IR<Code<Value>> visitBottom (IR<Code<Value>> ir) {
                        return ir;
                    }
                });
        }
        public separable <U> Code<Value> tryCastCode (DSLType<U> tp, U c) {
            return tp.acceptCode (c, new TypeVisitorCode<Code<Value>>() {
                    public separable Code<Value> visitTop (AnyCode c) {
                        return c.cast (Value.class);
                    }
                    public separable Code<Value> visitInteger (Code<Integer> c) {
                        return intCodeToValue (c);
                    }
                    public separable Code<Value> visitBoolean (Code<Boolean> c) {
                        return booleanCodeToValue (c);
                    }
                    public separable Code<Value> visitArray (Code<Value[]> c) {
                        return arrayCodeToValue (c);
                    }
                    public separable Code<Value> visitDouble (Code<Double> c) {
                        return doubleCodeToValue (c);
                    }
                    public separable Code<Value> visitBottom (Code<Value> c) {
                        return c;
                    }
                });
        }
        public <U> JoinReturn<?> join (final IR<Code<Value>> ir1, IR<U> ir2) {
            return ir2.tp.accept (ir2, new TypeVisitor<JoinReturn<?>>() {
                    public JoinReturn<?> visitTop (IR<AnyCode> ir) {
                        return new JoinReturn<Code<Value>>
                            (new DSLTypeBottom (),
                             ir1, new BoxAnyVal (ir));
                    }
                    public JoinReturn<?> visitInteger (IR<Code<Integer>> ir) {
                        return new JoinReturn<Code<Value>>
                            (new DSLTypeBottom (), ir1, new BoxIntValue (ir));
                    }
                    public JoinReturn<?> visitBoolean (IR<Code<Boolean>> ir) {
                        return new JoinReturn<Code<Value>>
                            (new DSLTypeBottom (), ir1, new BoxBoolValue (ir));
                    }
                    public JoinReturn<?> visitArray (IR<Code<Value[]>> ir) {
                        return new JoinReturn<Code<Value>>
                            (new DSLTypeBottom (), ir1, new BoxArrayValue (ir));
                    }
                    public JoinReturn<?> visitDouble (IR<Code<Double>> ir) {
                        return new JoinReturn<Code<Value>>
                            (new DSLTypeBottom (), ir1, new BoxDoubleValue (ir));
                    }
                    public JoinReturn<?> visitBottom (IR<Code<Value>> ir) {
                        return new JoinReturn<Code<Value>>
                            (new DSLTypeBottom (), ir1, ir);
                    }
                });
        }
        public separable Code<Boolean> valueCmp (Code<Value> c1,
                                                 Code<Value> c2) {
            return <| (`c1).compare (`c2) |>;
        }
    }


    /**
     * typed environments, which map variable names to DSLTypes
     */

    public static interface TEnv {
        public DSLType<?> get(String y);
    }
    public static final TEnv tenv0 = new TEnv() {
            public DSLType<?> get(String s) {
                throw new EnvironmentException(s);
            }
        };

    public static <T> TEnv text(final TEnv env, final String x,
                                          final DSLType<T> elem) {

        return new TEnv() {
            public DSLType<?> get(String y) {
                // error: if (x.equals(y))
                if (x==y)
                    return elem;
                else
                    return env.get(y);
            }
        };
    }

    /**
     * A new version of typed function environments
     */

    public static class DSLFunType<Tint,Tbool,Tarray,Tdouble,Tvalue> {
        DSLType<Tint> int_tp;
        DSLType<Tbool> bool_tp;
        DSLType<Tarray> array_tp;
        DSLType<Tdouble> double_tp;
        DSLType<Tvalue> value_tp;
        public separable DSLFunType (DSLType<Tint> int_tp,
                                     DSLType<Tbool> bool_tp,
                                     DSLType<Tarray> array_tp,
                                     DSLType<Tdouble> double_tp,
                                     DSLType<Tvalue> value_tp) {
            this.int_tp = int_tp;
            this.bool_tp = bool_tp;
            this.array_tp = array_tp;
            this.double_tp = double_tp;
            this.value_tp = value_tp;
        }
        public separable static <Tint,Tbool,Tarray,Tdouble,Tvalue> DSLFunType<Tint,Tbool,Tarray,Tdouble,Tvalue>
            make (DSLType<Tint> int_tp, DSLType<Tbool> bool_tp, DSLType<Tarray> array_tp,
                  DSLType<Tdouble> double_tp, DSLType<Tvalue> value_tp) {
            return new DSLFunType<Tint,Tbool,Tarray,Tdouble,Tvalue>(int_tp, bool_tp, array_tp, double_tp, value_tp);
        }
    }

    public static interface TFEnv {
        public <T> DSLType<?> get(String s, DSLType<T> arg_tp);
    }
    public static final TFEnv tfenv0 = new TFEnv() {
            public <T> DSLType<?> get(String s, DSLType<T> arg_tp) {
                throw new EnvironmentException(s);
            }
        };

    public static TFEnv tfext(final TFEnv fenv, final String x,
                              final DSLFunType<?,?,?,?,?> fun_tp) {
        return new TFEnv() {
            public <T> DSLType<?> get(String y, DSLType<T> arg_tp) {
                // error: if (x.equals(y))
                if (x==y) {
                    return arg_tp.accept (null, new TypeVisitor<DSLType<?>> () {
                            public DSLType<?> visitTop (IR<AnyCode> ir) {
                                return fun_tp.int_tp;
                            }
                            public DSLType<?> visitInteger (IR<Code<Integer>> ir) {
                                return fun_tp.int_tp;
                            }
                            public DSLType<?> visitBoolean (IR<Code<Boolean>> ir) {
                                return fun_tp.bool_tp;
                            }
                            public DSLType<?> visitArray (IR<Code<Value[]>> ir) {
                                return fun_tp.array_tp;
                            }
                            public DSLType<?> visitDouble (IR<Code<Double>> ir) {
                                return fun_tp.double_tp;
                            }
                            public DSLType<?> visitBottom (IR<Code<Value>> ir) {
                                return fun_tp.value_tp;
                            }
                        });
                } else
                    return fenv.get(y, arg_tp);
            }
        };
    }


    /**
     * IR<T> is the intermediate representation of expressions that can
     * produce a T, where T is generally a type of the form Code<T'>.
     */

    public static abstract class IR<T> {
        public DSLType<T> tp;
        private RWSet rwSet;
        public separable abstract T eval(Env e, FEnv f);
        IR (DSLType<T> tp, RWSet rwSet) { this.tp = tp; this.rwSet = rwSet; }
        public <U> IR<U> tryCast (DSLType<U> tp_to, boolean unboxOnFailure) {
            return tp_to.tryCast (this, unboxOnFailure);
        }
        public separable RWSet rwSet () { return rwSet; }
    }


    /**
     * The IR nodes for boxing, i.e., downcasting
     */

    public static abstract class BoxIR<T,U> extends IR<U> {
        IR<T> subNode;
        BoxIR (IR<T> subNode, DSLType<U> tp) {
            super (tp, subNode.rwSet());
            this.subNode = subNode; this.tp = tp;
        }
        public separable abstract U boxCode (T c);
        public separable U eval(Env e, FEnv f) {
            return boxCode (subNode.eval (e, f));
        }        
    }

    public static class BoxAnyInt extends BoxIR<AnyCode,Code<Integer>> {
        BoxAnyInt (IR<AnyCode> subNode) {
            super (subNode, new DSLTypeInteger ());
        }
        public separable Code<Integer> boxCode (AnyCode c) {
            return c.cast (Integer.class);
        }
    }

    public static class BoxAnyBool extends BoxIR<AnyCode,Code<Boolean>> {
        BoxAnyBool (IR<AnyCode> subNode) {
            super (subNode, new DSLTypeBoolean ());
        }
        public separable Code<Boolean> boxCode (AnyCode c) {
            return c.cast (Boolean.class);
        }
    }

    public static class BoxAnyArray extends BoxIR<AnyCode,Code<Value[]>> {
        BoxAnyArray (IR<AnyCode> subNode) {
            super (subNode, new DSLTypeArray ());
        }
        public separable Code<Value[]> boxCode (AnyCode c) {
            return c.cast (Value[].class);
        }
    }

    public static class BoxAnyDouble extends BoxIR<AnyCode,Code<Double>> {
        BoxAnyDouble (IR<AnyCode> subNode) {
            super (subNode, new DSLTypeDouble ());
        }
        public separable Code<Double> boxCode (AnyCode c) {
            return c.cast (Double.class);
        }
    }

    public static class BoxIntValue extends BoxIR<Code<Integer>,Code<Value>> {
        BoxIntValue (IR<Code<Integer>> subNode) {
            super (subNode, new DSLTypeBottom ());
        }
        public separable Code<Value> boxCode (Code<Integer> c) {
            return <| (Value) new IntValue (`c) |>;
        }
    }

    public static class BoxIntDouble extends BoxIR<Code<Integer>,Code<Double>> {
        BoxIntDouble (IR<Code<Integer>> subNode) {
            super (subNode, new DSLTypeDouble ());
        }
        public separable Code<Double> boxCode (Code<Integer> c) {
            return <| (double) (`c) |>;
        }
    }

    public static class BoxBoolValue extends BoxIR<Code<Boolean>,Code<Value>> {
        BoxBoolValue (IR<Code<Boolean>> subNode) {
            super (subNode, new DSLTypeBottom ());
        }
        public separable Code<Value> boxCode (Code<Boolean> c) {
            return <| (Value) new BooleanValue (`c) |>;
        }
    }

    public static class BoxArrayValue extends BoxIR<Code<Value[]>,Code<Value>> {
        BoxArrayValue (IR<Code<Value[]>> subNode) {
            super (subNode, new DSLTypeBottom ());
        }
        public separable Code<Value> boxCode (Code<Value[]> c) {
            return <| (Value) new ArrayValue (`c) |>;
        }
    }

    public static class BoxDoubleValue extends BoxIR<Code<Double>,Code<Value>> {
        BoxDoubleValue (IR<Code<Double>> subNode) {
            super (subNode, new DSLTypeBottom ());
        }
        public separable Code<Value> boxCode (Code<Double> c) {
            return <| (Value) new DoubleValue (`c) |>;
        }
    }

    public static class BoxAnyVal extends BoxIR<AnyCode,Code<Value>> {
        BoxAnyVal (IR<AnyCode> subNode) {
            super (subNode, new DSLTypeBottom ());
        }
        public separable Code<Value> boxCode (AnyCode c) {
            return c.cast (Value.class);
        }
    }


    /**
     * IR nodes for unboxing, i.e., conditionally projecting a Value
     */

    public static abstract class UnboxIR<T,U> extends IR<T> {
        IR<U> subNode;
        UnboxIR (DSLType<T> tp, IR<U> subNode) {
            super (tp, subNode.rwSet());
            this.subNode = subNode;
        }
        public separable abstract T unboxCode (U c);
        public separable T eval(Env e, FEnv f) {
            return unboxCode (subNode.eval (e, f));
        }        
    }

    public static class UnboxIntIR extends UnboxIR<Code<Integer>,Code<Value>> {
        UnboxIntIR (IR<Code<Value>> subNode) {
            super (new DSLTypeInteger (), subNode);
        }
        public separable Code<Integer> unboxCode (Code<Value> c) {
            return <| (`c).intValue () |>;
        }
    }

    public static class UnboxBoolIR extends UnboxIR<Code<Boolean>,Code<Value>> {
        UnboxBoolIR (IR<Code<Value>> subNode) {
            super (new DSLTypeBoolean (), subNode);
        }
        public separable Code<Boolean> unboxCode (Code<Value> c) {
            return <| (`c).booleanValue () |>;
        }
    }

    public static class UnboxArrayIR extends UnboxIR<Code<Value[]>,Code<Value>> {
        UnboxArrayIR (IR<Code<Value>> subNode) {
            super (new DSLTypeArray (), subNode);
        }
        public separable Code<Value[]> unboxCode (Code<Value> c) {
            return <| (`c).arrayValue () |>;
        }
    }

    public static class UnboxDoubleIR extends UnboxIR<Code<Double>,Code<Value>> {
        UnboxDoubleIR (IR<Code<Value>> subNode) {
            super (new DSLTypeDouble (), subNode);
        }
        public separable Code<Double> unboxCode (Code<Value> c) {
            return <| (`c).doubleValue () |>;
        }
    }

    // for when unboxing fails
    public static class ErrorIRAny<U> extends UnboxIR<AnyCode,Code<U>> {
        final Class<U> Uclass;
        ErrorIRAny (IR<Code<U>> subNode, Class<U> Uclass) {
            super (new DSLTypeTop (), subNode);
            this.Uclass = Uclass;
        }
        public separable AnyCode unboxCode (final Code<U> c) {
            return new AnyCode () {
                public separable <X> Code<X> cast (final Class<X> clazz) {
                    return <| let U unused = `c; WrongTypeException.<X>returningThrow() |>;
                }
            };
        }
    }

    public static class ErrorIRCode<T,U> extends UnboxIR<Code<T>,Code<U>> {
        final Class<T> Tclass;
        ErrorIRCode (DSLType<Code<T>> tp, Class<T> Tclass, IR<Code<U>> subNode) {
            super (tp, subNode);
            this.Tclass = Tclass;
        }
        public static separable <T,U> Code<T> helper (final Class<T> Tclass,
                                                      Code<U> c) {
            return <| let Object unused = `c; WrongTypeException.<T>returningThrow () |>;
        }
        public separable Code<T> unboxCode (Code<U> c) {
            return helper (Tclass, c);
        }
    }


    /**
     * IR nodes corresponding to the untyped expressions
     */

    public static class ValIR<T> extends IR<T> {
        private SValue<T> svalue;
        public ValIR(SValue<T> svalue) {
            super (svalue.tp, new RWSet ());
            this.svalue = svalue;
        }
        public static <T> ValIR<T> make (SValue<T> svalue) {
            return new ValIR<T> (svalue);
        }
        public separable T eval(Env e, FEnv f) {
            return svalue.code;
        }
    }
    
    public static class VarIR<T> extends IR<T> {
        private String s;
        public VarIR(DSLType<T> tp, String s) {
            super (tp, new RWSet ());
            this.s = s; // we assume s has already been interned
            //System.out.println ("making var " + s + " of type " + tp);
        }
        public static <T> VarIR<T> make (DSLType<T> tp, String s) {
            return new VarIR<T> (tp, s);
        }
        public separable T eval(Env e, FEnv f) {
            return e.get(s, this.tp);
        }
    }

    public static class AppIR<T,Targ> extends IR<T> {
        private String s;
        private IR<Targ> arg;
        public AppIR(String s, DSLType<T> tp, IR<Targ> arg) {
            super (tp, new RWSet (true));
            this.s = s; // we assume s has already been interned
            this.arg = arg;
        }
        public static <T,Targ> AppIR<T,Targ> make (String s, DSLType<T> tp,
                                                   IR<Targ> arg) {
            return new AppIR<T,Targ>(s, tp, arg);
        }
        public separable T eval(Env e, FEnv f) {
            return f.apply(s, tp, arg.tp, arg.eval(e,f));
        }
    }


    public static class BinOpIR<T,Tleft,Tright> extends IR<T> {
        protected IR<Tleft> left;
        protected IR<Tright> right;
        protected BinOpFun<T,Tleft,Tright> op_fun;
        public BinOpIR(IR<Tleft> left, IR<Tright> right,
                       BinOpFun<T,Tleft,Tright> op_fun) {
            super (op_fun.tp_out, left.rwSet().union (right.rwSet ()));
            this.left = left;
            this.right = right;
            this.op_fun = op_fun;
        }
        public static <Tout,Tleft,Tright,TIRl,TIRr> IR<Tout>
            make (BinOpFun<Tout,Tleft,Tright> op,
                  IR<TIRl> ir_left, IR<TIRr> ir_right) {
            IR<Tleft> left = op.tp_l.tryCast (ir_left, true);
            IR<Tright> right = op.tp_r.tryCast (ir_right, true);
            return new BinOpIR<Tout,Tleft,Tright> (left, right, op);
        }
        public separable T eval (Env e, FEnv f) {
            return op_fun.apply (left.eval (e, f), right.eval (e, f));
        }
    }


    /**
     * abstract binary operations
     */

    public static abstract class BinOpFun<Tout,Tl,Tr> {
        public DSLType<Tout> tp_out;
        public DSLType<Tl> tp_l;
        public DSLType<Tr> tp_r;
        public BinOpFun (DSLType<Tout> tp_out, DSLType<Tl> tp_l,
                         DSLType<Tr> tp_r) {
            this.tp_out = tp_out;
            this.tp_l = tp_l;
            this.tp_r = tp_r;
        }
        public separable abstract Tout apply (Tl c_left, Tr c_right);
    }

    public static abstract class IntBinOpFun extends BinOpFun<Code<Integer>,Code<Integer>,Code<Integer>> {
        public IntBinOpFun () {
            super (new DSLTypeInteger (), new DSLTypeInteger (),
                   new DSLTypeInteger ());
        }
    }

    public static abstract class DoubleBinOpFun extends BinOpFun<Code<Double>,Code<Double>,Code<Double>> {
        public DoubleBinOpFun () {
            super (new DSLTypeDouble (), new DSLTypeDouble (),
                   new DSLTypeDouble ());
        }
    }

    public static abstract class BoolTBinOpFun<T> extends BinOpFun<Code<Boolean>,T,T> {
        public BoolTBinOpFun (DSLType<T> in_tp) {
            super (new DSLTypeBoolean (), in_tp, in_tp);
        }
    }

    public static DoubleBinOpFun addOp = new DoubleBinOpFun () {
            public separable Code<Double> apply (Code<Double> c_left,
                                                 Code<Double> c_right) {
                return <| `c_left + `c_right |>;
            }
        };

    public static IntBinOpFun addOpInt = new IntBinOpFun () {
            public separable Code<Integer> apply (Code<Integer> c_left,
                                                  Code<Integer> c_right) {
                return <| `c_left + `c_right |>;
            }
        };

    public static DoubleBinOpFun subOp = new DoubleBinOpFun () {
            public separable Code<Double> apply (Code<Double> c_left,
                                                 Code<Double> c_right) {
                return <| `c_left - `c_right |>;
            }
        };

    public static IntBinOpFun subOpInt = new IntBinOpFun () {
            public separable Code<Integer> apply (Code<Integer> c_left,
                                                  Code<Integer> c_right) {
                return <| `c_left - `c_right |>;
            }
        };

    public static DoubleBinOpFun mulOp = new DoubleBinOpFun () {
            public separable Code<Double> apply (Code<Double> c_left,
                                                 Code<Double> c_right) {
                return <| `c_left * `c_right |>;
            }
        };

    public static IntBinOpFun mulOpInt = new IntBinOpFun () {
            public separable Code<Integer> apply (Code<Integer> c_left,
                                                  Code<Integer> c_right) {
                return <| `c_left * `c_right |>;
            }
        };

    public static DoubleBinOpFun divOp = new DoubleBinOpFun () {
            public separable Code<Double> apply (Code<Double> c_left,
                                                 Code<Double> c_right) {
                return <| `c_left / `c_right |>;
            }
        };

    public static IntBinOpFun divOpInt = new IntBinOpFun () {
            public separable Code<Integer> apply (Code<Integer> c_left,
                                                  Code<Integer> c_right) {
                return <| `c_left / `c_right |>;
            }
        };

    public static BoolTBinOpFun<Code<Double>> lessOp =
        new BoolTBinOpFun<Code<Double>> (new DSLTypeDouble ()) {
            public separable Code<Boolean> apply (Code<Double> c_left,
                                                  Code<Double> c_right) {
                return <| `c_left < `c_right |>;
            }
    };

    public static BoolTBinOpFun<Code<Boolean>> andOp =
        new BoolTBinOpFun<Code<Boolean>> (new DSLTypeBoolean ()) {
            public separable Code<Boolean> apply (Code<Boolean> c_left,
                                                  Code<Boolean> c_right) {
                return <| `c_left && `c_right |>;
            }
    };


    /**
     * Equality is somewhat special: if Tleft and Tright are comparable,
     * then we downcast both to the most general type and compare them,
     * otherwise we evaluate both sides (to make sure we get the effects) and
     * just return false; we represent these two possibilities with different
     * subclaasses of EqualsIR, and makeEqualsIR decides which one to use
     */
    public static class EqualsIR<Tleft,Tright> extends BinOpIR<Code<Boolean>,Tleft,Tright> {
        public EqualsIR(IR<Tleft> left, IR<Tright> right,
                        BinOpFun<Code<Boolean>,Tleft,Tright> op_fun) {
            super(left, right, op_fun);
        }

        public static <Tleft,Tright> EqualsIR<?,?> make (IR<Tleft> l,
                                                         IR<Tright> r) {
            IR<Tleft> boxed_r = l.tp.tryCast (r, false);
            IR<Tright> boxed_l = r.tp.tryCast (l, false);
            if (boxed_r != null) {
                return new EqualsIR<Tleft,Tleft>
                    (l, boxed_r, new EqualsCmpOpFun<Tleft> (l.tp));
            } else if (boxed_l != null) {
                return new EqualsIR<Tright,Tright>
                    (boxed_l, r, new EqualsCmpOpFun<Tright> (r.tp));
            } else {
                return make (toCodeIR (l), toCodeIR (r));
            }
        }

        // this helper make instantiates the type variables...
        public static <Tl,Tr> EqualsIR<Code<Tl>,Code<Tr>> make (CodeIR<Tl> l,
                                                                CodeIR<Tr> r) {
            return new EqualsIR<Code<Tl>,Code<Tr>>
                (l.ir, r.ir, new EqualsNoCmpOpFun<Tl,Tr> (l.ir.tp, l.clazz,
                                                          r.ir.tp, r.clazz));
        }
    }

    public static class EqualsCmpOpFun<T> extends BoolTBinOpFun<T> {
        public EqualsCmpOpFun(DSLType<T> tp) {
            super(tp);
        }
        public separable Code<Boolean> apply (T c_left, T c_right) {
            return tp_l.valueCmp (c_left, c_right);
        }
    }

    // README: EqualsNoCmpOpFun only works for code types
    public static class EqualsNoCmpOpFun<Tleft,Tright> extends BinOpFun<Code<Boolean>,Code<Tleft>,Code<Tright>> {
        Class<Tleft> class_left;
        Class<Tright> class_right;
        public EqualsNoCmpOpFun(DSLType<Code<Tleft>> tp_l,
                                Class<Tleft> class_left,
                                DSLType<Code<Tright>> tp_r,
                                Class<Tright> class_right) {
            super(new DSLTypeBoolean (), tp_l, tp_r);
            this.class_left = class_left; this.class_right = class_right;
        }
        // need a helper so we can have variables of type Class<Tleft> and
        // Class<Tright> in scope
        public static separable <Tl,Tr> Code<Boolean> helper (Code<Tl> c_left,
                                                              final Class<Tl> class_l,
                                                              Code<Tr> c_right,
                                                              final Class<Tr> class_r) {
            return <| let Tl left = `c_left; let Tr right = `c_right; false |>;
        }
        public separable Code<Boolean> apply (Code<Tleft> c_left,
                                              Code<Tright> c_right) {
            return helper (c_left, class_left, c_right, class_right);
        }
    }

    // helper for EqualsIRNoCmp: converts an IR<T> to a pair of a Class<U>
    // and an IR<Code<U>> for some U
    public static class CodeIR<T> {
        public Class<T> clazz;
        public IR<Code<T>> ir;
        CodeIR (Class<T> clazz, IR<Code<T>> ir) {
            this.clazz = clazz; this.ir = ir;
        }
    }
    public static <T> CodeIR<?> toCodeIR (IR<T> ir) {
        return ir.tp.accept (ir, new TypeVisitor<CodeIR<?>>() {
                public CodeIR<?> visitTop (IR<AnyCode> ir) {
                    return new CodeIR<Integer> (Integer.class,
                                                new BoxAnyInt (ir));
                }
                public CodeIR<?> visitInteger (IR<Code<Integer>> ir) {
                    return new CodeIR<Integer> (Integer.class, ir);
                }
                public CodeIR<?> visitBoolean (IR<Code<Boolean>> ir) {
                    return new CodeIR<Boolean> (Boolean.class, ir);
                }
                public CodeIR<?> visitArray (IR<Code<Value[]>> ir) {
                    return new CodeIR<Value[]> (Value[].class, ir);
                }
                public CodeIR<?> visitDouble (IR<Code<Double>> ir) {
                    return new CodeIR<Double> (Double.class, ir);
                }
                public CodeIR<?> visitBottom (IR<Code<Value>> ir) {
                    return new CodeIR<Value> (Value.class, ir);
                }
            });
    }

    public static class NotIR extends IR<Code<Boolean>> {
        private IR<Code<Boolean>> subNode;
        public NotIR(IR<Code<Boolean>> subNode) {
            super (new DSLTypeBoolean (), subNode.rwSet());
            this.subNode = subNode;
        }
        public separable Code<Boolean> eval(Env e, FEnv f) {
            return <| !`(subNode.eval(e,f)) |>;
        }
    }

    public static class IfIR<T> extends IR<Code<T>> {
        private IR<Code<Boolean>> test;
        private IR<Code<T>> conseq, alt;
        private final Class<T> clazz;
        public IfIR (IR<Code<Boolean>> test, Class<T> clazz,
                     IR<Code<T>> conseq, IR<Code<T>> alt) {
            super (conseq.tp, test.rwSet().union (conseq.rwSet ().union (alt.rwSet ())));
            this.test = test;
            this.clazz = clazz;
            this.conseq = conseq;
            this.alt = alt;
        }
        // each of these instantiates another type variable, so they
        // must be separate methods...
        public static <T> IfIR<T> make (IR<Code<Boolean>> test,
                                        CodeIR2<T> code_irs) {
            return new IfIR<T> (test, code_irs.clazz, code_irs.ir1, code_irs.ir2);
        }
        public static <T1,T2,T3> IfIR<?> make (IR<T1> test, IR<T2> conseq, IR<T3> alt) {
            return make (test.tryCast (new DSLTypeBoolean (), true),
                         joinToCodeIR2 (conseq.tp.join (conseq, alt)));
        }
        public separable Code<T> eval(Env e, FEnv f) {
            return <| `(test.eval(e,f)) ? `(conseq.eval (e, f)) : `(alt.eval (e, f)) |>;
        }
    }

    public static class LetIR<T,U> extends IR<Code<T>> {
        private String _var;
        private IR<Code<U>> _rhs;
        private final Class<U> _uClazz;
        private IR<Code<T>> _body;
        private final Class<T> _tClazz;
        public LetIR (String var, IR<Code<U>> rhs, Class<U> uClazz, IR<Code<T>> body, Class<T> tClazz) {
            super (body.tp, ((IR)rhs instanceof AllocArrayIR)?
                       rhs.rwSet().union(body.rwSet()).removeR(var).removeW(var):
                       rhs.rwSet().union(body.rwSet().completeIfContains(var)));
            _var = var;
            _rhs = rhs;
            _uClazz = uClazz;
            _body = body;
            _tClazz = tClazz;
        }
        public separable Code<T> eval(final Env e, final FEnv f) {
            // Naive let
//            Env newenv = ext (e, _var, _rhs.tp, _rhs.eval (e, f));
//            return _body.eval (newenv, f);

            final Code<U> rhsVal = _rhs.eval (e, f);
            final String fvar = _var;
            final IR<Code<T>> fbody = _body;
            return _rhs.tp.acceptCode (rhsVal, new TypeVisitorCode<Code<T>>() {
                    public separable Code<T> visitTop (AnyCode c) {
                        return <| let final int temp = `(c.cast (Integer.class)); // FIXME: is this correct?
                        `(fbody.eval (ext (e, fvar, new DSLTypeInteger(), <| temp |>), f))|>;
                    }
                    public separable Code<T> visitInteger (Code<Integer> c) {
                        return <| let final int temp = `(c);
                        `(fbody.eval (ext (e, fvar, new DSLTypeInteger(), <| temp |>), f))|>;
                    }
                    public separable Code<T> visitBoolean (Code<Boolean> c) {
                        return <| let final boolean temp = `(c);
                        `(fbody.eval (ext (e, fvar, new DSLTypeBoolean(), <| temp |>), f))|>;
                    }
                    public separable Code<T> visitArray (Code<Value[]> c) {
                        return <| let final Value[] temp = `(c);
                        `(fbody.eval (ext (e, fvar, new DSLTypeArray(), <| temp |>), f))|>;
                    }
                    public separable Code<T> visitDouble (Code<Double> c) {
                        return <| let final double temp = `(c);
                        `(fbody.eval (ext (e, fvar, new DSLTypeDouble(), <| temp |>), f))|>;
                    }
                    public separable Code<T> visitBottom (Code<Value> c) {
                        return <| let final Value temp = `(c);
                        `(fbody.eval (ext (e, fvar, new DSLTypeBottom(), <| temp |>), f))|>;
                    }
                });
        }
        public static <T,U> LetIR<T,U> make (String var, CodeIR<U> rhs, CodeIR<T> body) {
            return new LetIR<T,U> (var, rhs.ir, rhs.clazz, body.ir, body.clazz);
        }
        public static <T1,T2> LetIR<?,?> make (String var, IR<T1> rhs, IR<T2> body) {
            return make (var, toCodeIR(rhs), toCodeIR(body));
        }
    }

    public static class AllocArrayIR extends IR<Code<Value[]>> {
        private IR<Code<Integer>> _size;
        private IR<Code<Value>> _elem;
        public AllocArrayIR(IR<Code<Integer>> size, IR<Code<Value>> elem) {
            super (new DSLTypeArray (), size.rwSet().union (elem.rwSet ()));
            _size = size;
            _elem = elem;
        }
        public separable Code<Value[]> eval(Env e, FEnv f) {
            return <| allocArray (`(_size.eval (e, f)),
                                  `(_elem.eval (e, f))) |>;
        }
    }

    public static class AGetIR extends IR<Code<Value>> {
        private IR<Code<Value[]>> _arr;
        private IR<Code<Integer>> _index;
        public AGetIR(IR<Code<Value[]>> arr, IR<Code<Integer>> index) {
            super (new DSLTypeBottom (),
                   (arr instanceof VarIR)?
                       (index.rwSet().addR (((VarIR)arr).s)):
                       (index.rwSet().completeR ()));
            _arr = arr;
            _index = index;
        }
        public separable Code<Value> eval(Env e, FEnv f) {
            return <| `(_arr.eval (e, f))[`(_index.eval (e, f))] |>;
        }
    }
    
    public static class ASetIR extends IR<Code<Integer>> {
        private IR<Code<Value[]>> _arr;
        private IR<Code<Integer>> _index;
        private IR<Code<Value>> _elem;
        public ASetIR(IR<Code<Value[]>> arr, IR<Code<Integer>> index, IR<Code<Value>> elem) {
            super (new DSLTypeInteger (),
                   (arr instanceof VarIR)?
                       (index.rwSet().union(elem.rwSet()).addW (((VarIR)arr).s)):
                       (index.rwSet().union(elem.rwSet()).completeW ()));
            _arr = arr;
            _index = index;
            _elem = elem;
        }
        public separable Code<Integer> eval(Env e, FEnv f) {
            return <| setArray(`(_arr.eval (e, f)), `(_index.eval (e, f)), `(_elem.eval (e, f))) |>;
        }
    }    
    
    public static Value[] allocArray (int size, Value elem) {
        Value [] arr = new Value [size];
        for (int i = 0; i < arr.length; ++i) { arr[i] = elem; }
        return arr;
    }

    public static int setArray (Value[] arr, int i, Value elem) {
        arr[i] = elem;
        return 0;
    }
    
    public static class FloorIR extends IR<Code<Integer>> {
        private IR<Code<Double>> subNode;
        public FloorIR(IR<Code<Double>> subNode) {
            super (new DSLTypeInteger (), subNode.rwSet());
            this.subNode = subNode;
        }
        public separable Code<Integer> eval(Env e, FEnv f) {
            return <| ((Double)(`(subNode.eval(e,f)))).intValue() |>;
        }
    }

    /** Random number 0 <= x < 1 */
    public static class RandIR extends IR<Code<Double>> {
        public RandIR() {
            super (new DSLTypeDouble (), new RWSet());
        }
        public separable Code<Double> eval(Env e, FEnv f) {
            return <| getRandom() |>;
        }
    }
    
    public static final HashMap<Thread,Random> RANDOM_FOR_THREAD = new HashMap<Thread,Random>();
    
    public static double getRandom() {
        Random r = RANDOM_FOR_THREAD.get(Thread.currentThread());
        if (r == null) {
            RANDOM_FOR_THREAD.put(Thread.currentThread(), r = new Random()); // seed?
        }
        return r.nextDouble();
    }    
    
    public static class SForIR extends IR<Code<Integer>> {
        protected String _var;
        protected IR<Code<Integer>> _bound;
        protected IR<Code<Value>> _body;
        public SForIR(String var, IR<Code<Integer>> bound, IR<Code<Value>> body) {
            super (new DSLTypeInteger (), bound.rwSet().union (body.rwSet ().completeIfContains (var)));
            _var = var.intern(); // must intern here, so we can use == in the fext method below;
            _bound = bound;
            _body = body;
        }
        
        public separable Code<Integer> eval(final Env e, final FEnv f) {
            // FIXME: mybody should not need to be final...
            final IR<Code<Value>> mybody = _body;
            final String myvar = _var;
            return <|
                let LoopRunner runner = new LoopRunner() {
                        public int run (int lower, int upper) {
                            for (int i = lower; i < upper; ++i) {
                                final int fi = i;
                                Value unused = `(mybody.eval (ext (e, myvar, new DSLTypeInteger (), <| fi |>), f));
                            }
                            return 0;
                        }};
                let int bound = `(_bound.eval (e, f));
                runner.run (0, bound) |>;
        }
    }
    
    public static class ForIR extends SForIR {
        private boolean _runParallel;
        private String[] rSet, wSet;
        public ForIR(String var, IR<Code<Integer>> bound, IR<Code<Value>> body, boolean runParallel) {
            super (var, bound, body);
            rSet = new String[body.rwSet().readElems().size()];
            int i = 0;
            for(String rVar: body.rwSet().readElems()) rSet[i++] = rVar;
            wSet = new String[body.rwSet().writeElems().size()];
            i = 0;
            for(String wVar: body.rwSet().writeElems()) wSet[i++] = wVar;
            // System.out.println("runParallel = "+runParallel+
            //                    ", rSet = "+java.util.Arrays.toString(rSet)+
            //                    ", wSet = "+java.util.Arrays.toString(wSet));
            _runParallel = runParallel;
        }
        
        public separable Code<Integer> eval(final Env e, final FEnv f) {
            if (!_runParallel) return super.eval(e, f);
            
            // FIXME: mybody should not need to be final...
            final IR<Code<Value>> mybody = _body;
            final String myvar = _var;
            return <|
                let LoopRunner runner = new LoopRunner() {
                        public int run (int lower, int upper) {
                            for (int i = lower; i < upper; ++i) {
                                final int fi = i;
                                Value unused = `(mybody.eval (ext (e, myvar, new DSLTypeInteger (), <| fi |>), f));
                            }
                            return 0;
                        }};
                let int bound = `(_bound.eval (e, f));
                let boolean runParallel = (!`(rwOverlaps(e, rSet, wSet)) && poolThreadsAvailable());
                runParallel ?
                /*util.Misc.printAndReturn(*/runner.runParallel(bound)/*,"runParallel "+myvar)*/ :
                /*util.Misc.printAndReturn(*/runner.runSerial(bound)/*,"runSerial "+myvar)*/ |>;
        }
    }
   
    public abstract static class LoopRunner {
        public abstract int run (int lower, int upper);
        public int runSerial(final int bound) { return run(0, bound); }
        public int runParallel (final int bound) {
            final int numThreadsInThisLoop = Math.max(1,Math.min(pool().getMaximumPoolSize()-pool().getActiveCount(), bound));
            Collection<Callable<Integer>> tasks =
                new ArrayList<Callable<Integer>> (numThreadsInThisLoop);
            final int chunk_size = bound / numThreadsInThisLoop;
            final int additional_item = bound % numThreadsInThisLoop; // how many threads get one additional item
//            util.IO.addOutput("LoopRunner.runParallel(bound="+bound+"), numThreadsInThisLoop="+numThreadsInThisLoop+
//                              ", chunk_size="+chunk_size+", additional_item = "+additional_item+
//                              ": "+(additional_item)+"x"+(chunk_size+1)+" + "+(numThreadsInThisLoop-additional_item)+"x"+chunk_size);
            for (int i = 0; i < numThreadsInThisLoop; ++i) {
                final int ii = i;
                final LoopRunner tthis = this;
                final int from = (i < additional_item) 
                    ? i*(chunk_size+1) 
                    : additional_item*(chunk_size+1) + (i-additional_item)*chunk_size;
                final int to = (i < additional_item) 
                    ? (i+1)*(chunk_size+1)
                    : additional_item*(chunk_size+1) + (i-additional_item+1)*chunk_size;
                // util.IO.addOutput("Thread "+i+" --> from="+from+", to="+to);
                tasks.add (new Callable<Integer> () {
                        public Integer call () {
                            return tthis.run (from, to);
                        }});
            }
            try {
                List<Future<Integer>> futures = pool().invokeAll(tasks);
                for(Future<Integer> f: futures) {
                    f.get(); // this propagates exceptions out of the child threads
                }
            }
            catch(ExecutionException ee) { throw new EvaluationException(ee); }
            catch(InterruptedException ie) { throw new EvaluationException(ie); }
            return 0;
        }
    }
    
    public static boolean poolThreadsAvailable() {
        return (pool().getMaximumPoolSize()-pool().getActiveCount()>=NUM_THREADS);
    }
    
        
    /** Display an ArrayValue of DoubleValues as square grayscale image. */
    public static class DisplayImageIR extends IR<Code<Integer>> {
        IR<Code<Value[]>> _image;
        String _title;
        public DisplayImageIR(IR<Code<Value[]>> image, String title) {
            super(new DSLTypeInteger(), image.rwSet());
            _image = image; _title = title;
        }
        public separable Code<Integer> eval(Env e, FEnv f) {
            final IR<Code<Value[]>> fimage = _image;
            final String ftitle = _title;
            return <| displayImageHelper(`(_image.eval(e, f)), ftitle) |>;
        }
    }

    public static int displayImageHelper(Value[] a, String title) {
        if (a.length==0) { throw new EvaluationException("Empty image"); }
        int side = (int)Math.sqrt(a.length);
        if (side*side != a.length) {
            throw new EvaluationException("Not a square image: image.length="+a.length+", side="+side);
        }
        double[] image = new double[a.length];
        for(int i=0; i<a.length; ++i) {
            image[i] = Math.max(0, Math.min(1, a[i].doubleValue())); // clamp to 0<=temp<=1                
        }
        util.Misc.displayImages(side, title, image);
        return 0;
    }
    
    /** Load an image into a ArrayValue of DoubleValues. */
    public static class LoadImageIR extends IR<Code<Value[]>> {
        String _fileName;
        public LoadImageIR(String fileName) {
            super(new DSLTypeArray(), new RWSet());
            _fileName = fileName;
        }
        public separable Code<Value[]> eval(Env e, FEnv f) {
            final String ffileName = _fileName;
            return <| loadImageHelper(ffileName) |>;
        }
    }

    public static Value[] loadImageHelper(String fileName) {
        try {
            java.awt.image.BufferedImage image = javax.imageio.ImageIO.read(new java.io.File(fileName));
            int side = Math.min(image.getWidth(),image.getHeight());
            Value[] a = new Value[side*side];
            java.awt.image.Raster r = image.getData(new java.awt.Rectangle(0,0,side,side));
            int i=0;
            for(int y=0; y<side; ++y) {
                for(int x=0; x<side; ++x) {
                    double acc = 0;
                    for(int band=0; band<r.getNumBands(); ++band) {
                        acc += r.getSampleDouble(x, y, band) / Math.pow(2, r.getSampleModel().getSampleSize(band));
                    }
                    a[i++] = new DoubleValue(acc/r.getNumBands());
                }
            }
            return a;
        }
        catch(java.io.IOException ioe) { throw new EvaluationException(ioe); }
    }

    /** Make a Gaussian blur kernel with a given size. */
    public static class GaussianKernelIR extends IR<Code<Value[]>> {
        IR<Code<Integer>> _size;
        public GaussianKernelIR(IR<Code<Integer>> size) {
            super(new DSLTypeArray(), size.rwSet());
            _size = size;
        }
        public separable Code<Value[]> eval(Env e, FEnv f) {
            return <| gaussianKernelHelper(`(_size.eval(e, f))) |>;
        }
    }
    
    public static Value[] gaussianKernelHelper(int size) {
        double[] kernel = util.Misc.makeKernel(size);
        Value[] a = new Value[kernel.length];
        for(int i=0; i<kernel.length; ++i) {
            a[i] = new DoubleValue(kernel[i]);
        }
        return a;
    }

    // helper for IfIR: converts a JoinReturn<T> to a Class<U> and 2 IR<U>'s
    public static class CodeIR2<T> {
        public Class<T> clazz;
        public IR<Code<T>> ir1, ir2;
        CodeIR2 (Class<T> clazz, IR<Code<T>> ir1, IR<Code<T>> ir2) {
            this.clazz = clazz; this.ir1 = ir1; this.ir2 = ir2;
        }
    }
    public static <T> CodeIR2<?> joinToCodeIR2 (JoinReturn<T> joinRet) {
        return joinRet.ir1.tp.accept2
            (joinRet.ir1, joinRet.ir2,
             new TypeVisitor2<CodeIR2<?>>() {
                public CodeIR2<?> visitTop (IR<AnyCode> ir1, IR<AnyCode> ir2) {
                    return new CodeIR2<Integer> (Integer.class,
                                                 new BoxAnyInt (ir1),
                                                 new BoxAnyInt (ir2));
                }
                public CodeIR2<?> visitInteger (IR<Code<Integer>> ir1,
                                                IR<Code<Integer>> ir2) {
                    return new CodeIR2<Integer> (Integer.class, ir1, ir2);
                }
                public CodeIR2<?> visitBoolean (IR<Code<Boolean>> ir1,
                                                IR<Code<Boolean>> ir2) {
                    return new CodeIR2<Boolean> (Boolean.class, ir1, ir2);
                }
                public CodeIR2<?> visitArray (IR<Code<Value[]>> ir1,
                                              IR<Code<Value[]>> ir2) {
                    return new CodeIR2<Value[]> (Value[].class, ir1, ir2);
                }
                public CodeIR2<?> visitDouble (IR<Code<Double>> ir1,
                                               IR<Code<Double>> ir2) {
                    return new CodeIR2<Double> (Double.class, ir1, ir2);
                }
                public CodeIR2<?> visitBottom (IR<Code<Value>> ir1,
                                               IR<Code<Value>> ir2) {
                    return new CodeIR2<Value> (Value.class, ir1, ir2);
                }
            });
    }



    /**
     * untyped expressions
     */

    public static interface Exp {
        public IR<?> analyze (TEnv e, TFEnv f);
    }

    public static class Val implements Exp {
        private Value value;
        public Val(Value value) {
            this.value = value;
        }
        public IR<?> analyze (TEnv e, TFEnv f) {
            if (value instanceof IntValue) {
                return new ValIR<Code<Integer>> (SValue.make (value.intValue()));
            } else if (value instanceof BooleanValue) {
                return new ValIR<Code<Boolean>> (SValue.make (value.booleanValue()));
            } else if (value instanceof DoubleValue) {
                return new ValIR<Code<Double>> (SValue.make (value.doubleValue()));
            } else {
                throw new WrongTypeException();
            }
        }
    }
    
    public static class Var implements Exp {
        private String s;
        public Var(String s) {
            this.s = s.intern(); // must intern here, so we can use == below
        }
        public IR<?> analyze (TEnv e, TFEnv f) {
            return VarIR.make (e.get (s), s);
        }
    }
    
    public static class App implements Exp {
        private final String _s;
        private final Exp _arg;
        public App(String s, Exp arg) {
            _s = s.intern(); // must intern here, so we can use == in the ext method below
            _arg = arg;
        }
        public IR<?> analyze (TEnv e, TFEnv f) {
            IR<?> argIR = _arg.analyze (e, f);
            return AppIR.make (_s, f.get (_s, argIR.tp), argIR);
        }
    }
    
    public static abstract class BinOp implements Exp {
        protected Exp left, right;
        protected BinOpFun<?,?,?> op_fun;
        public BinOp(Exp left, Exp right, BinOpFun<?,?,?> op_fun) {
            this.left = left;
            this.right = right;
            this.op_fun = op_fun;
        }
        public IR<?> analyze (TEnv e, TFEnv f) {
            IR<?> l = left.analyze (e, f);
            IR<?> r = right.analyze (e, f);
            
            return BinOpIR.make (op_fun, l, r);
        }
    }

    public static abstract class DoubleIntBinOp extends BinOp {
        protected BinOpFun<?,?,?> op_fun_int;
        public DoubleIntBinOp(Exp left, Exp right, BinOpFun<?,?,?> op_fun, BinOpFun<?,?,?> op_fun_int) {
            super(left, right, op_fun);
            this.op_fun_int = op_fun_int;
        }
        public IR<?> analyze (TEnv e, TFEnv f) {
            IR<?> l = left.analyze (e, f);
            IR<?> r = right.analyze (e, f);
            IR<Code<Integer>> intl = l.tryCast (new DSLTypeInteger (), false);
            IR<Code<Integer>> intr = r.tryCast (new DSLTypeInteger (), false);
            if (intl != null && intr != null) {
                return BinOpIR.make (op_fun_int, intl, intr);
            }
            else {
                return BinOpIR.make (op_fun, l, r);
            }
        }
    }
    
    public static class Add extends DoubleIntBinOp {
        public Add(Exp left, Exp right) { super(left, right, addOp, addOpInt); }
    }
    
    public static class Sub extends DoubleIntBinOp {
        public Sub(Exp left, Exp right) { super(left, right, subOp, subOpInt); }
    }
    
    public static class Mul extends DoubleIntBinOp {
        public Mul(Exp left, Exp right) { super(left, right, mulOp, mulOpInt); }
    }
    
    public static class Div extends DoubleIntBinOp {
        public Div(Exp left, Exp right) { super(left, right, divOp, divOpInt); }
    }

    public static class Less extends BinOp {
        public Less(Exp left, Exp right) { super(left, right, lessOp); }
    }
    
    public static class And extends BinOp {
        public And(Exp left, Exp right) { super(left, right, andOp); }
    }
    
    public static class Not implements Exp {
        private Exp _exp;
        public Not(Exp exp) {
            _exp = exp;
        }
        public IR<?> analyze (TEnv e, TFEnv f) {
            return new NotIR (_exp.analyze (e, f).tryCast (new DSLTypeBoolean (), true));
        }
    }


    public static class Equals implements Exp {
        private Exp left, right;
        public Equals(Exp left, Exp right) {
            this.left = left; this.right = right;
        }
        public IR<?> analyze (TEnv e, TFEnv f) {
            return EqualsIR.make (left.analyze(e,f), right.analyze(e,f));
        }
    }


    public static class If implements Exp {
        private Exp test, conseq, alt;
        public If(Exp test, Exp conseq, Exp alt) {
            this.test = test;
            this.conseq = conseq;
            this.alt = alt;
        }
        public IR<?> analyze (TEnv e, TFEnv f) {
            return IfIR.make (test.analyze (e, f),
                              conseq.analyze (e, f), alt.analyze (e, f));
        }
    }

    public static class Let implements Exp {
        private String _var;
        private Exp _rhs, _body;
        public Let(String var, Exp rhs, Exp body) {
            _var = var.intern();
            _rhs = rhs;
            _body = body;
        }
        public IR<?> analyze (TEnv e, TFEnv f) {
            IR<?> rhs_ir = _rhs.analyze (e, f);
            return LetIR.make (_var,
                               rhs_ir,
                               _body.analyze (text (e, _var, rhs_ir.tp), f));
        }
    }
    
    public static class AllocArray implements Exp {
        private Exp _size, _elem;
        public AllocArray(Exp size, Exp elem) {
            _size = size;
            _elem = elem;
        }
        public IR<?> analyze(TEnv e, TFEnv f) {
            return new AllocArrayIR(_size.analyze(e,f).tryCast (new DSLTypeInteger (), true),
                                    _elem.analyze(e,f).tryCast (new DSLTypeBottom (), true));
        }
    }
    
    public static class AGet implements Exp {
        private Exp _arr, _index;
        public AGet(Exp arr, Exp index) {
            _arr = arr;
            _index = index;
        }
        public IR<?> analyze(TEnv e, TFEnv f) {
            return new AGetIR(_arr.analyze(e,f).tryCast (new DSLTypeArray (), true),
                              _index.analyze(e,f).tryCast (new DSLTypeInteger (), true));
        }
    }
    
    public static class ASet implements Exp {
        public Exp _arr, _index, _elem;
        public ASet(Exp arr, Exp index, Exp elem) {
            _arr = arr;
            _index = index;
            _elem = elem;
        }
        public IR<?> analyze(TEnv e, TFEnv f) {
            return new ASetIR(_arr.analyze(e,f).tryCast (new DSLTypeArray (), true),
                              _index.analyze(e,f).tryCast (new DSLTypeInteger (), true),
                              _elem.analyze(e,f).tryCast (new DSLTypeBottom (), true));
        }
    }
    
    public static class Rand implements Exp {
        public Rand() { }
        public IR<?> analyze (TEnv e, TFEnv f) {
            return new RandIR ();
        }
    }

    public static class Floor implements Exp {
        private Exp _exp;
        public Floor(Exp exp) {
            _exp = exp;
        }
        public IR<?> analyze (TEnv e, TFEnv f) {
            return new FloorIR (_exp.analyze (e, f).tryCast (new DSLTypeDouble (), true));
        }
    }
   
    public static class SFor implements Exp {
        protected String _var;
        protected Exp _bound, _body;
        public SFor(String var, Exp bound, Exp body) {
            _var = var;
            _bound = bound;
            _body = body;
        }
        public IR<?> analyze(TEnv e, TFEnv f) {
            return new SForIR(_var,
                              _bound.analyze(e,f).tryCast (new DSLTypeInteger (), true),
                              _body.analyze(text (e, _var, new DSLTypeInteger ()),f).
                                  tryCast (new DSLTypeBottom (), true));
        }
    }
    
    public static class For extends SFor {
        public For(String var, Exp bound, Exp body) { super(var, bound, body); }
        public IR<?> analyze(TEnv e, TFEnv f) {
            IR<Code<Value>> bodyIR = _body.analyze(text (e, _var, new DSLTypeInteger ()),f).
                tryCast (new DSLTypeBottom (), true);
            boolean runParallel = !bodyIR.rwSet().overlaps();
            return new ForIR(_var,
                             _bound.analyze(e,f).tryCast (new DSLTypeInteger (), true),
                             bodyIR, runParallel);
        }
    }

    // for computing if a read and write set overlap
    public static separable Code<Boolean> rwOverlaps (Env e, String[] rSet, String[] wSet) {
        Code<Boolean> c = <| false |>;
        for (String rVar : rSet) {
            Code<Value> r = e.get (rVar, new DSLTypeBottom());
            for (String wVar : wSet) {
                Code<Value> w = e.get (wVar, new DSLTypeBottom());
                c = <| `c || (`r instanceof ArrayValue
                                  && `w instanceof ArrayValue
                                  && ((`r).arrayValue () == (`w).arrayValue ())) |>;
            }
        }
        return c;
    }
    
    /** Display an ArrayValue of DoubleValues as square grayscale image. */
    public static class DisplayImage implements Exp {
        Exp _image;
        String _title;
        public DisplayImage(Exp image, String title) {
            _image = image; _title = title;
        }
        public IR<?> analyze(TEnv e, TFEnv f) {
            return new DisplayImageIR(_image.analyze(e,f).tryCast (new DSLTypeArray (), true),
                                      _title);
        }
    }

    /** Load an image into a ArrayValue of DoubleValues. */
    public static class LoadImage implements Exp {
        String _fileName;
        public LoadImage(String fileName) { _fileName = fileName; }
        public IR<?> analyze(TEnv e, TFEnv f) { return new LoadImageIR(_fileName); }
    }

    /** Make a Gaussian blur kernel with a given size. */
    public static class GaussianKernel implements Exp {
        Exp _size;
        public GaussianKernel(Exp size) { _size = size; }
        public IR<?> analyze(TEnv e, TFEnv f) {
            return new GaussianKernelIR(_size.analyze(e,f).tryCast (new DSLTypeInteger (), true));
        }
    }

    /**
     * code environments
     */

    // environment; implemented as a function object from String to SValue
    public static interface Env {
        public separable <T> T get(String y, DSLType<T> tp);
    }
    public static final Env env0 = new Env() {
            public separable <T> T get(String s, DSLType<T> tp) {
                throw new EnvironmentException(s);
            }
        };

    public static separable <T> Env ext(final Env env,
                                        final String x,
                                        final DSLType<T> tp,
                                        final T c) {
        return new Env() {
            public separable <U> U get(String y, DSLType<U> req_tp) {
                // error: if (x.equals(y))
                if (x==y) {
                    U ret = req_tp.tryCastCode (tp, c);
                    if (ret == null)
                        throw new WrongTypeException ();
                    else
                        return ret;
                } else
                    return env.get(y, req_tp);
            }
        };
    }

    /**
     * functions and function environments
     */

    // interface to represent Value -> Value functions
    public static interface ValueFun<Tint,Tbool,Tarray,Tdouble,Tvalue> {
        public Tint applyInt (final int x);
        public Tbool applyBool (final boolean x);
        public Tarray applyArray (final Value[] x);
        public Tdouble applyDouble (final double x);
        public Tvalue applyValue (final Value x);
    }

    public static interface SValueFun {
        public separable <T> SValue<?> apply (SValue<T> x);
    }

    public separable static <Ti,Tb,Ta,Td,Tv> SValueFun makeSValueFun (final Code<ValueFun<Ti,Tb,Ta,Td,Tv>> fun,
                                                                      final DSLFunType<Code<Ti>,Code<Tb>,Code<Ta>,Code<Td>,Code<Tv>> fun_tp,
                                                                      final Class<Ti> ti,
                                                                      final Class<Tb> tb,
                                                                      final Class<Ta> ta,
                                                                      final Class<Td> td,
                                                                      final Class<Tv> tv) {
        return new SValueFun () {
            public separable <T> SValue<?> apply (SValue<T> x) {
                    return x.tp.acceptCode
                        (x.code, new TypeVisitorCode<SValue<?>>() {
                            public separable SValue<?> visitTop (AnyCode c) {
                                final Class<Ti> ti2 = ti;
                                return new SValue<Code<Ti>>
                                (fun_tp.int_tp,
                                 <| `(fun).applyInt (`(c.cast (Integer.class))) |>);
                            }
                            public separable SValue<?> visitInteger (Code<Integer> c) {
                                final Class<Ti> ti2 = ti;
                                return new SValue<Code<Ti>>
                                (fun_tp.int_tp, <| `(fun).applyInt (`c) |>);
                            }
                            public separable SValue<?> visitBoolean (Code<Boolean> c) {
                                final Class<Tb> tb2 = tb;
                                return new SValue<Code<Tb>>
                                (fun_tp.bool_tp, <| `(fun).applyBool (`c) |>);
                            }
                            public separable SValue<?> visitArray (Code<Value[]> c) {
                                final Class<Ta> ta2 = ta;
                                return new SValue<Code<Ta>>
                                (fun_tp.array_tp, <| `(fun).applyArray (`c) |>);
                            }
                            public separable SValue<?> visitDouble (Code<Double> c) {
                                final Class<Td> td2 = td;
                                return new SValue<Code<Td>>
                                (fun_tp.double_tp, <| `(fun).applyDouble (`c) |>);
                            }
                            public separable SValue<?> visitBottom (Code<Value> c) {
                                final Class<Tv> tv2 = tv;
                                return new SValue<Code<Tv>>
                                (fun_tp.value_tp, <| `(fun).applyValue (`c) |>);
                            }
                        });
            }
        };
    }


    // function environment; implemented as a function object from String to Code<SValueFun>
    public static interface FEnv {
        public separable <Tin,Tout> Tout apply(String s, DSLType<Tout> tp_out,
                                               DSLType<Tin> tp_in, Tin arg);
    }
    public static final FEnv fenv0 = new FEnv() {
        public separable <Tin,Tout> Tout apply(String s, DSLType<Tout> tp_out,
                                               DSLType<Tin> tp_in, Tin arg) {
            throw new EnvironmentException(s);
        }
    };

    public static separable FEnv fext(final FEnv fenv,
                                      final String x,
                                      final SValueFun fun) {
        return new FEnv() {
            public separable <Tin,Tout> Tout apply(String s, DSLType<Tout> tp_out,
                                                   DSLType<Tin> tp_in, Tin arg) {
                if (x==s) {
                    SValue<?> output =
                        fun.apply (new SValue<Tin>(tp_in, arg));
                    Tout retval = tp_out.tryCastCode (output);
                    if (retval != null) {
                        return retval;
                    } else {
                        throw new WrongTypeException ();
                    }
                } else
                    return fenv.apply(s, tp_out, tp_in, arg);
            }
        };
    }


    public static separable <Ti,Tb,Ta,Td,Tv> FEnv
        fext(FEnv fenv, String x,
             Code<ValueFun<Ti,Tb,Ta,Td,Tv>> fun,
             DSLFunType<Code<Ti>,Code<Tb>,Code<Ta>,Code<Td>,Code<Tv>> fun_tp,
             Class<Ti> ti, Class<Tb> tb, Class<Ta> ta,
             Class<Td> td, Class<Tv> tv) {
        return fext (fenv, x, makeSValueFun (fun, fun_tp, ti, tb, ta, td, tv));
    }


    /**
     *  README: below is the "safer" version of SValueFun and fext
     */

    /*
    // interface to represent functions on code types
    public static interface SValueFun<Tint,Tbool,Tvalue> {
        DSLFunType<Ti,Tb,Tv> fun_tp;
        public separable Tint applyInt (Code<Integer> x);
        public separable Tbool applyBool (Code<Boolean> x);
        public separable Tvalue applyValue (Code<Value> x);
    }

    // apply Code<ValueFun> to an argument
    public static separable <Ti,Tb,Tv,Tin> SValue<?>
        applyValueFunCode (final DSLFunType<Ti,Tb,Tv> fun_tp,
                           final SValueFun<Ti,Tb,Tv> fun,
                           DSLType<Tin> tp_in, Tin arg) {
        return tp_in.acceptCode
            (arg,
             new TypeVisitorCode<SValue<?>> () {
                public separable SValue<?> visitTop (AnyCode c) {
                    return new SValue<Code<Ti>>
                        (fun_tp.int_tp,
                         <| `(fun).applyInt (`(c.cast (Integer.class))) |>);
                }
                public separable SValue<?> visitInteger (Code<Integer> c) {
                    return new SValue<Code<Ti>>
                        (fun_tp.int_tp,
                         <| `(fun).applyInt (`c) |>);
                }
                public separable SValue<?> visitBoolean (Code<Boolean> c) {
                    return toCode (fun_tp.bool_tp,
                                   <| `(fun).applyBool (`c) |>);
                }
                public separable SValue<?> visitBottom (Code<Value> c) {
                    return toCode (fun_tp.value_tp,
                                   <| `(fun).applyValue (`c) |>);
                }
            });
    }


    public static separable <Ti,Tb,Tv> FEnv fext(final FEnv fenv,
                                                 final String x,
                                                 final DSLFunType<Ti,Tb,Tv> fun_tp,
                                                 final Code<ValueFun<Ti,Tb,Tv>> fun) {
        return new FEnv() {
            public separable <Tin,Tout> Tout apply(String s, DSLType<Tout> tp_out,
                                                   Tin arg, DSLType<Tin> tp_in) {
                if (x==s) {
                    SValue<?> output =
                        applyValueFunCode (fun_tp, fun, tp_in, arg);
                    Tout retval = tp_out.tryCastCode (output.tp, output.code);
                    if (retval != null) {
                        return retval;
                    } else {
                        throw new WrongTypeException ();
                    }
                } else
                    return fenv.apply(s, tp_out, tp_in, arg);
            }
        };
    }

    */


    /**
     * A FunIR<Code<Tint,Tbool>> is IR for a function that returns type Tint
     * for integer arguments and type Tbool for boolean arguments.
     */

    public static class FunIR<Tint,Tbool,Tarray,Tdouble,Tvalue> {
        public IR<Tint> intIR;
        public IR<Tbool> boolIR;
        public IR<Tarray> arrayIR;
        public IR<Tdouble> doubleIR;
        public DSLType<Tvalue> value_tp;
        public FunIR (IR<Tint> intIR, IR<Tbool> boolIR, IR<Tarray> arrayIR,
                      IR<Tdouble> doubleIR, DSLType<Tvalue> value_tp) {
            this.intIR = intIR; this.boolIR = boolIR; this.arrayIR = arrayIR; 
            this.doubleIR = doubleIR; this.value_tp = value_tp;
        }
        public static <Tint,Tbool,Tarray,Tdouble,Tvalue> FunIR<Tint,Tbool,Tarray,Tdouble,Tvalue>
            tryMake (DSLType<Tint> int_tp, DSLType<Tbool> bool_tp,
                     DSLType<Tarray> array_tp, DSLType<Tdouble> double_tp,
                     DSLType<Tvalue> value_tp, IR<?> intIR, IR<?> boolIR,
                     IR<?> arrayIR, IR<?> doubleIR) {
            IR<Tint> castedIntIR = int_tp.tryCast (intIR, false);
            IR<Tbool> castedBoolIR = bool_tp.tryCast (boolIR, false);
            IR<Tarray> castedArrayIR = array_tp.tryCast (arrayIR, false);
            IR<Tdouble> castedDoubleIR = double_tp.tryCast (doubleIR, false);
            IR<Tvalue> castedIntValueIR = value_tp.tryCast (intIR, false);
            IR<Tvalue> castedBoolValueIR = value_tp.tryCast (boolIR, false);
            IR<Tvalue> castedArrayValueIR = value_tp.tryCast (arrayIR, false);
            IR<Tvalue> castedDoubleValueIR = value_tp.tryCast (doubleIR, false);
            if (castedIntIR != null && castedBoolIR != null
                && castedArrayIR != null && castedDoubleIR != null
                && castedIntValueIR != null && castedBoolValueIR != null
                && castedArrayValueIR != null && castedDoubleValueIR != null) {
                return new FunIR<Tint,Tbool,Tarray,Tdouble,Tvalue>
                    (castedIntIR, castedBoolIR, castedArrayIR, castedDoubleIR,
                     value_tp);
            } else {
                return null;
            }
        }
    }


    /**
     * We make a FunIR by doing iterated dataflow: first assume that the
     * return types for int, bool and array are top, and then iteratively refine
     * it until we reach a fixed point.
     */

    public static FunIR<?,?,?,?,?> makeFunIR (TEnv tenv, TFEnv tfenv, Exp exp,
                                              String fname, String param) {
        DSLType<?> int_tp = new DSLTypeTop ();
        DSLType<?> bool_tp = new DSLTypeTop ();
        DSLType<?> array_tp = new DSLTypeTop ();
        DSLType<?> double_tp = new DSLTypeTop ();
        DSLType<?> value_tp = new DSLTypeTop ();
        FunIR<?,?,?,?,?> retval = null;
        while (retval == null) {
            TFEnv tfenv_ext =
                tfext (tfenv, fname,
                       DSLFunType.make (int_tp, bool_tp, array_tp, double_tp, value_tp));
            IR<?> intIR = exp.analyze (text (tenv, param, new DSLTypeInteger ()), tfenv_ext);
            IR<?> boolIR = exp.analyze (text (tenv, param, new DSLTypeBoolean ()), tfenv_ext);
            IR<?> arrayIR = exp.analyze (text (tenv, param, new DSLTypeArray ()), tfenv_ext);
            IR<?> doubleIR = exp.analyze (text (tenv, param, new DSLTypeDouble ()), tfenv_ext);
            retval = FunIR.tryMake (int_tp, bool_tp, array_tp, double_tp, value_tp,
                                    intIR, boolIR, arrayIR, doubleIR);
            // update our tps, in case tryMake failed
            int_tp = intIR.tp;
            bool_tp = boolIR.tp;
            array_tp = arrayIR.tp;
            double_tp = doubleIR.tp;
            value_tp = joinHelper (joinHelper (joinHelper (intIR, boolIR), arrayIR), doubleIR).tp;
        }
        return retval;
    }

    public static <T,U> IR<?> joinHelper (IR<T> ir1, IR<U> ir2) {
        return ir1.tp.join (ir1, ir2).ir1;
    }


    /**
     * IR for a function that is known to return Code types
     */

    // stupid Java: we can't just re-use toCodeIR below because Java
    // complains that ? is not equal to ?
    public static class DSLCodeType<T> {
        public DSLType<Code<T>> tp;
        public Class<T> clazz;
        DSLCodeType (DSLType<Code<T>> tp, Class<T> clazz) {
            this.tp = tp; this.clazz = clazz;
        }
    }
    // BUG: we should just be able to call toCodeIR here...
    public static <T> DSLCodeType<?> toCodeTp (DSLType<T> tp) {
        return tp.accept (null, new TypeVisitor<DSLCodeType<?>>() {
                public DSLCodeType<?> visitTop (IR<AnyCode> ir) {
                    return new DSLCodeType<Integer> (new DSLTypeInteger (), // MGR: why Integer?
                                                     Integer.class);
                }
                public DSLCodeType<?> visitInteger (IR<Code<Integer>> ir) {
                    return new DSLCodeType<Integer> (new DSLTypeInteger (),
                                                     Integer.class);
                }
                public DSLCodeType<?> visitBoolean (IR<Code<Boolean>> ir) {
                    return new DSLCodeType<Boolean> (new DSLTypeBoolean (),
                                                     Boolean.class);
                }
                public DSLCodeType<?> visitArray (IR<Code<Value[]>> ir) {
                    return new DSLCodeType<Value[]> (new DSLTypeArray (),
                                                     Value[].class);
                }
                public DSLCodeType<?> visitDouble (IR<Code<Double>> ir) {
                    return new DSLCodeType<Double> (new DSLTypeDouble (),
                                                    Double.class);
                }
                public DSLCodeType<?> visitBottom (IR<Code<Value>> ir) {
                    return new DSLCodeType<Value> (new DSLTypeBottom (),
                                                     Value.class);
                }
            });
    }

    // BUG: we shouldn't need to define this extra class...
    public static class CodeFunIR<Ti,Tb,Ta,Td,Tv> {
        FunIR<Code<Ti>,Code<Tb>,Code<Ta>,Code<Td>,Code<Tv>> fun_ir;
        Class<Ti> ti_class;
        Class<Tb> tb_class;
        Class<Ta> ta_class;
        Class<Td> td_class;
        Class<Tv> tv_class;
        CodeFunIR (FunIR<Code<Ti>,Code<Tb>,Code<Ta>,Code<Td>,Code<Tv>> fun_ir,
                   Class<Ti> ti_class, Class<Tb> tb_class, Class<Ta> ta_class,
                   Class<Td> td_class, Class<Tv> tv_class) {
            this.fun_ir = fun_ir;
            this.ti_class = ti_class;
            this.tb_class = tb_class;
            this.ta_class = ta_class;
            this.td_class = td_class;
            this.tv_class = tv_class;
        }
        public separable DSLFunType<Code<Ti>,Code<Tb>,Code<Ta>,Code<Td>,Code<Tv>> getDSLFunType () {
            return new DSLFunType<Code<Ti>,Code<Tb>,Code<Ta>,Code<Td>,Code<Tv>>
                (fun_ir.intIR.tp,
                 fun_ir.boolIR.tp,
                 fun_ir.arrayIR.tp,
                 fun_ir.doubleIR.tp,
                 fun_ir.value_tp);
        }
        public static <Tint,Tbool,Tarray,Tdouble,Tvalue> CodeFunIR<Tint,Tbool,Tarray,Tdouble,Tvalue>
            tryMake (DSLCodeType<Tint> int_tp, DSLCodeType<Tbool> bool_tp,
                     DSLCodeType<Tarray> array_tp, DSLCodeType<Tdouble> double_tp, 
                     DSLCodeType<Tvalue> value_tp,
                     IR<?> intIR, IR<?> boolIR, IR<?> arrayIR, IR<?> doubleIR) {

            return new CodeFunIR<Tint,Tbool,Tarray,Tdouble,Tvalue>
                (FunIR.tryMake (int_tp.tp, bool_tp.tp, array_tp.tp, double_tp.tp, value_tp.tp,
                                intIR, boolIR, arrayIR, doubleIR),
                 int_tp.clazz, bool_tp.clazz, array_tp.clazz, double_tp.clazz, value_tp.clazz);
        }        

        public separable Code<ValueFun<Ti,Tb,Ta,Td,Tv>> eval (Env e, FEnv f, String fname, String param) {
            return evalFun (this, e, f, fname, param,
                            ti_class, tb_class, ta_class, td_class, tv_class);
        }

        public separable static <Ti,Tb,Ta,Td,Tv> Code<ValueFun<Ti,Tb,Ta,Td,Tv>>
            evalFun (final CodeFunIR<Ti,Tb,Ta,Td,Tv> fun, final Env e, final FEnv f,
                     final String fname, final String param,
                     final Class<Ti> ti, final Class<Tb> tb,
                     final Class<Ta> ta, final Class<Td> td, final Class<Tv> tv) {
            final Class<Ti> ti2 = ti;
            final Class<Tb> tb2 = tb;
            final Class<Ta> ta2 = ta;
            final Class<Td> td2 = td;
            final Class<Tv> tv2 = tv;
            return <| (ValueFun<Ti,Tb,Ta,Td,Tv>) new ValueFun<Ti,Tb,Ta,Td,Tv>() {
                public Ti applyInt (final int x) {
                    final ValueFun<Ti,Tb,Ta,Td,Tv> tthis = this;
                    return `(fun.fun_ir.intIR.eval
                             (ext (e, param, new DSLTypeInteger (), <| x |>),
                              fext (f, fname, <| tthis |>, fun.getDSLFunType (),
                                    ti, tb, ta, td, tv)));
                }
                public Tb applyBool (final boolean x) {
                    final ValueFun<Ti,Tb,Ta,Td,Tv> tthis = this;
                    return `(fun.fun_ir.boolIR.eval
                             (ext (e, param, new DSLTypeBoolean (), <| x |>),
                              fext (f, fname, <| tthis |>, fun.getDSLFunType (),
                                    ti, tb, ta, td, tv)));
                }
                public Ta applyArray (final Value[] x) {
                    final ValueFun<Ti,Tb,Ta,Td,Tv> tthis = this;
                    return `(fun.fun_ir.arrayIR.eval
                             (ext (e, param, new DSLTypeArray (), <| x |>),
                              fext (f, fname, <| tthis |>, fun.getDSLFunType (),
                                    ti, tb, ta, td, tv)));
                }
                public Td applyDouble (final double x) {
                    final ValueFun<Ti,Tb,Ta,Td,Tv> tthis = this;
                    return `(fun.fun_ir.doubleIR.eval
                             (ext (e, param, new DSLTypeDouble (), <| x |>),
                              fext (f, fname, <| tthis |>, fun.getDSLFunType (),
                                    ti, tb, ta, td, tv)));
                }
                public Tv applyValue (final Value x) {
                    final ValueFun<Ti,Tb,Ta,Td,Tv> tthis = this;
                    if (x instanceof IntValue) {
                        return `(castCodeCheck
                                 (fun.fun_ir.value_tp, fun.fun_ir.intIR.tp,
                                  <| tthis.applyInt (x.intValue ()) |>));
                    } else if (x instanceof BooleanValue) {
                        return `(castCodeCheck
                                 (fun.fun_ir.value_tp, fun.fun_ir.boolIR.tp,
                                  <| tthis.applyBool (x.booleanValue ()) |>));
                    } else if (x instanceof DoubleValue) {
                        return `(castCodeCheck
                                 (fun.fun_ir.value_tp, fun.fun_ir.doubleIR.tp,
                                  <| tthis.applyDouble (x.doubleValue ()) |>));
                    } else if (x instanceof DoubleValue) {
                        return `(castCodeCheck
                                 (fun.fun_ir.value_tp, fun.fun_ir.arrayIR.tp,
                                  <| tthis.applyArray (x.arrayValue ()) |>));
                    } else {
                        throw new WrongTypeException ();
                    }
                }
            } |>;
        }
    }

    public static separable <T,U> T castCodeCheck (DSLType<T> tp,
                                                   DSLType<U> tp_from, U c) {
        T ret = tp.tryCastCode (tp_from, c);
        if (ret == null) {
            util.IO.addOutput ("Could not cast from type " + tp_from
                               + " to type " + tp);
            throw new WrongTypeException ();
        }
        return ret;
    }


    public static CodeFunIR<?,?,?,?,?> makeCodeFunIR (TEnv tenv,
                                                      TFEnv tfenv,
                                                      Exp exp,
                                                      String fname,
                                                      String param) {
        // first find a fixed point with makeFunIR
        FunIR<?,?,?,?,?> tmp = makeFunIR (tenv, tfenv, exp, fname, param);

        // next we weaken all the types to be Code<?> types
        DSLCodeType<?> int_tp = toCodeTp(tmp.intIR.tp);
        DSLCodeType<?> bool_tp = toCodeTp(tmp.boolIR.tp);
        DSLCodeType<?> array_tp = toCodeTp(tmp.arrayIR.tp);
        DSLCodeType<?> double_tp = toCodeTp(tmp.doubleIR.tp);
        DSLCodeType<?> value_tp = toCodeTp(tmp.value_tp);
        CodeFunIR<?,?,?,?,?> retval = null;

        // and we find a fixed point again
        while (retval == null) {
            TFEnv tfenv_ext =
                tfext (tfenv, fname, DSLFunType.make (int_tp.tp, bool_tp.tp, array_tp.tp, double_tp.tp, value_tp.tp));
            CodeIR<?> intIR = toCodeIR (exp.analyze (text (tenv, param, new DSLTypeInteger ()), tfenv_ext));
            CodeIR<?> boolIR = toCodeIR (exp.analyze (text (tenv, param, new DSLTypeBoolean ()), tfenv_ext));
            CodeIR<?> arrayIR = toCodeIR (exp.analyze (text (tenv, param, new DSLTypeArray ()), tfenv_ext));
            CodeIR<?> doubleIR = toCodeIR (exp.analyze (text (tenv, param, new DSLTypeDouble ()), tfenv_ext));
            retval = CodeFunIR.tryMake (int_tp, bool_tp, array_tp, double_tp,
                                        value_tp, intIR.ir, boolIR.ir,
                                        arrayIR.ir, doubleIR.ir);
            // update our tps, in case tryMake failed
            int_tp = toCodeTp(intIR.ir.tp);
            bool_tp = toCodeTp(boolIR.ir.tp);
            array_tp = toCodeTp(arrayIR.ir.tp);
            double_tp = toCodeTp(doubleIR.ir.tp);
            value_tp = toCodeTp (joinHelper (joinHelper (joinHelper (intIR.ir, boolIR.ir), arrayIR.ir), doubleIR.ir).tp);
        }
        return retval;
    }


    /**
     * IR for whole programs
     */

    public static abstract class ProgramIR {
        public separable Code<Value> eval (FEnv f) { return eval(env0, f); }
        public separable abstract Code<Value> eval (Env e, FEnv f);
    }

    public static class BaseProgramIR extends ProgramIR {
        IR<Code<Value>> body;
        BaseProgramIR (IR<Code<Value>> body) { this.body = body; }
        public separable Code<Value> eval (Env e, FEnv f) {
            return body.eval (e, f);
        }
    }

    public static class BindProgramIR<Ti,Tb,Ta,Td,Tv> extends ProgramIR {
        CodeFunIR<Ti,Tb,Ta,Td,Tv> code_fun;
        ProgramIR body;
        String fname, param;
        BindProgramIR (CodeFunIR<Ti,Tb,Ta,Td,Tv> code_fun, ProgramIR body,
                       String fname, String param) {
            this.code_fun = code_fun; this.body = body;
            this.fname = fname; this.param = param;
        }
        public static <Ti,Tb,Ta,Td,Tv> BindProgramIR<Ti,Tb,Ta,Td,Tv>
            make (CodeFunIR<Ti,Tb,Ta,Td,Tv> code_fun, ProgramIR body,
                  String fname, String param) {
            return new BindProgramIR<Ti,Tb,Ta,Td,Tv> (code_fun, body, fname, param);
        }
        public separable Code<Value> eval (Env e, FEnv f) {
            return eval_helper (code_fun, e, f, body, fname, param,
                                code_fun.ti_class, code_fun.tb_class,
                                code_fun.ta_class, code_fun.td_class, 
                                code_fun.tv_class);
        }
        public static separable <Ti2,Tb2,Ta2,Td2,Tv2> Code<Value>
            eval_helper (CodeFunIR<Ti2,Tb2,Ta2,Td2,Tv2> code_fun2,
                         Env e, FEnv f, ProgramIR body2, String fname2, String param2,
                         final Class<Ti2> ti, final Class<Tb2> tb,
                         final Class<Ta2> ta, final Class<Td2> td, 
                         final Class<Tv2> tv) {
            // BUG: mint currently does not see ti, tb, ta, and tv above...
            final Class<Ti2> ti2 = ti;
            final Class<Tb2> tb2 = tb;
            final Class<Ta2> ta2 = ta;
            final Class<Td2> td2 = td;
            final Class<Tv2> tv2 = tv;

            return <| let ValueFun<Ti2,Tb2,Ta2,Td2,Tv2> fun = `(code_fun2.eval (e, f, fname2, param2));
            `(body2.eval (e, fext (f, fname2,
                                makeSValueFun (<| fun |>,
                                               DSLFunType.make (code_fun2.fun_ir.intIR.tp,
                                                                code_fun2.fun_ir.boolIR.tp,
                                                                code_fun2.fun_ir.arrayIR.tp,
                                                                code_fun2.fun_ir.doubleIR.tp,
                                                                code_fun2.fun_ir.value_tp),
                                               ti, tb, ta, td, tv)))) |>;
        }
    }


    /**
     * Declarations and programs
     */

    public static class Declaration {
        public String fun, param;
        public Exp body;
        public Declaration(String fun, String param, Exp body) {
            this.fun = fun.intern(); // must intern here, so we can use == in the ext method above
            this.param = param.intern(); // must intern here, so we can use == in the ext method above
            this.body = body;
        }
    }

    public static class Program {
        private Declaration[] _defs;
        private Exp _body;
        public Program(Exp body, Declaration ... defs) {
            _defs = defs;
            _body = body;
        }
        public Declaration[] defs() { return _defs; }
        public Exp body() { return _body; }
        
        public Code<Value> peval(TFEnv tfenv, FEnv fenv) {
            return analyze (tfenv, 0).eval (fenv);
        }

        public ProgramIR analyze(TEnv tenv, TFEnv tfenv) {
            return analyze(tenv, tfenv, 0);
        }
            
        private ProgramIR analyze(TFEnv tfenv, int defIndex) {
            return analyze(tenv0, tfenv, defIndex);
        }
        
        private ProgramIR analyze(TEnv tenv, TFEnv tfenv, int defIndex) {
            if (_defs.length<=defIndex) {
                return new BaseProgramIR
                    (new DSLTypeBottom().tryCast (_body.analyze(tenv,tfenv), true));
            }
            else {
                Declaration d = _defs[defIndex];
                CodeFunIR<?,?,?,?,?> fun_ir = makeCodeFunIR (tenv, tfenv, d.body,
                                                             d.fun, d.param);
                TFEnv new_tfenv = tfext (tfenv, d.fun, fun_ir.getDSLFunType ());
                ProgramIR rest = analyze (tenv, new_tfenv, defIndex + 1);
                return BindProgramIR.make (fun_ir, rest, d.fun, d.param);
            }
        }
    }
    
    public static ProgramIR bindFunction(String varName, String funName, String paramName, Exp functionBody) {
        return new Program
            (new App(funName, new Var(varName)),
             new Declaration(funName, paramName, functionBody)).
            analyze(text (tenv0, varName, new DSLTypeBottom()), tfenv0);
    }

    public static separable Code<Value> evalFunction(String varName, Code<Value> v, ProgramIR pir) {
        return pir.eval(ext(env0, varName, new DSLTypeBottom(), v), fenv0);
    }
    
    public static final parser.LintParser<Exp> PARSER = new parser.LintParser<Exp>(Exp.class)
        .addSubParser(new parser.LintParser<Value>(Value.class));

    public static Program parsedTermFact = new Program
        (PARSER.parse("(App f (Val (IntValue 10)))"),
         new Declaration
             ("f", "x", PARSER.parse("(If (Equals (Var x) (Val (IntValue 0)))"+
                                     "    (Val (IntValue 1))"+
                                     "    (Mul (Var x)"+
                                     "         (App f (Sub (Var x)"+
                                     "                     (Val (IntValue 1))))))")));

    public static Program parsedTermFib = new Program
        (PARSER.parse("(App f (Val (IntValue 10)))"),
         new Declaration
             ("f", "x", PARSER.parse("(If (Equals (Var x) (Val (IntValue 0)))"+
                                     "    (Val (IntValue 0))"+
                                     "    (If (Equals (Sub (Var x) (Val (IntValue 1)))"+
                                     "                (Val (IntValue 0)))"+
                                     "        (Val (IntValue 1))"+
                                     "        (Add (App f (Sub (Var x) (Val (IntValue 1))))"+
                                     "             (App f (Sub (Var x) (Val (IntValue 2)))))))")));
    
    public static Program parsedTermDoublePlusInt = new Program
        (PARSER.parse("(App f (Val (DoubleValue 10)))"),
         new Declaration
             ("f", "x", PARSER.parse("(Add (Var x) (Val (IntValue 1)))")));
    public static Program parsedTermIntPlusInt = new Program
        (PARSER.parse("(App f (Val (IntValue 10)))"),
         new Declaration
             ("f", "x", PARSER.parse("(Add (Var x) (Val (IntValue 1)))")));
    public static Program parsedTermDoubleEqualsInt = new Program
        (PARSER.parse("(App f (Val (DoubleValue 1)))"),
         new Declaration
             ("f", "x", PARSER.parse("(Equals (Var x) (Val (IntValue 1)))")));
    public static Program parsedTermIntEqualsInt = new Program
        (PARSER.parse("(App f (Val (IntValue 1)))"),
         new Declaration
             ("f", "x", PARSER.parse("(Equals (Var x) (Val (IntValue 1)))")));
    public static Program parsedTermDoubleLessInt = new Program
        (PARSER.parse("(App f (Val (DoubleValue 0)))"),
         new Declaration
             ("f", "x", PARSER.parse("(Less (Var x) (Val (IntValue 1)))")));
    public static Program parsedTermIntLessInt = new Program
        (PARSER.parse("(App f (Val (IntValue 0)))"),
         new Declaration
             ("f", "x", PARSER.parse("(Less (Var x) (Val (IntValue 1)))")));
    public static Program parsedTermLet = new Program
        (PARSER.parse("(App f (Val (IntValue 123)))"),
         new Declaration
             ("f", "x", PARSER.parse("(Let y (Var x) (Var y))")));
    public static Program parsedTermAllocArray = new Program
        (PARSER.parse("(App f (Val (IntValue 10)))"),
         new Declaration
             ("f", "x", PARSER.parse("(AllocArray (Var x) (Val (DoubleValue 3.14)))")));
    public static Program parsedTermAGet = new Program
        (PARSER.parse("(App f (Val (IntValue 1)))"),
         new Declaration
             ("f", "x", PARSER.parse("(AGet (AllocArray (Val (IntValue 10)) (Val (DoubleValue 3.14))) (Var x))")));
    public static Program parsedTermASet = new Program
        (PARSER.parse("(App f (Val (IntValue 1)))"),
         new Declaration
             ("f", "x", PARSER.parse("(ASet (AllocArray (Val (IntValue 10)) (Val (DoubleValue 3.14))) (Var x) (Var x))")));
    public static Program parsedTermASet2 = new Program
        (PARSER.parse("(App f (Val (IntValue 1)))"),
         new Declaration
             ("f", "x", PARSER.parse("(Let a (AllocArray (Val (IntValue 10)) (Val (DoubleValue 3.14)))"+
                                     "     (Let dummy (ASet (Var a) (Var x) (Var x))"+
                                     "          (Var a)))")));
    public static Program parsedTermRand = new Program
        (PARSER.parse("(App f (Val (IntValue 1)))"),
         new Declaration
             ("f", "x", PARSER.parse("(Rand)")));
    public static Program parsedTermFloor = new Program
        (PARSER.parse("(App f (Val (DoubleValue 3.14)))"),
         new Declaration
             ("f", "x", PARSER.parse("(Floor (Var x))")));
    public static Program parsedTermFloorInt = new Program
        (PARSER.parse("(App f (Val (IntValue 3)))"),
         new Declaration
             ("f", "x", PARSER.parse("(Floor (Var x))")));
    public static Program parsedTermSFor = new Program
        (PARSER.parse("(App f (Val (IntValue 10)))"),
         new Declaration
             ("f", "x", PARSER.parse("(Let a (AllocArray (Var x) (Val (IntValue 0)))"+
                                     "     (Let dummy (SFor i (Var x) (ASet (Var a) (Var i) (Var i)))"+
                                     "          (Var a)))")));
    public static Program parsedTermFor = new Program
        (PARSER.parse("(App f (Val (IntValue 10)))"),
         new Declaration
             ("f", "x", PARSER.parse("(Let a (AllocArray (Var x) (Val (IntValue 0)))"+
                                     "     (Let dummy (For i (Var x) (ASet (Var a) (Var i) (Var i)))"+
                                     "          (Var a)))")));
    public static Program parsedTermForNotParallel = new Program
        (PARSER.parse("(App f (Val (IntValue 10)))"),
         new Declaration
             ("f", "x", PARSER.parse("(Let a (AllocArray (Var x) (Val (IntValue 2)))"+
                                     "     (Let dummy (For i (Sub (Var x) (Val (IntValue 1))) "+
                                     "                       (ASet (Var a) "+
                                     "                             (Add (Var i) (Val (IntValue 1))) "+
                                     "                             (AGet (Var a) (Var i))))"+
                                     "          (Var a)))")));
    public static Program parsedTermForNotParallel2 = new Program
        (PARSER.parse("(App f (Val (IntValue 10)))"),
         new Declaration
             ("f", "x", PARSER.parse("(Let a (AllocArray (Var x) (Val (IntValue 2)))"+
                                     "     (Let b (Var a)"+
                                     "            (Let dummy (For i (Sub (Var x) (Val (IntValue 1))) "+
                                     "                              (ASet (Var a) "+
                                     "                                    (Add (Var i) (Val (IntValue 1))) "+
                                     "                                    (AGet (Var b) (Var i))))"+
                                     "                 (Var a))))")));
    public static Program parsedTermGaussianKernel = new Program
        (PARSER.parse("(App f (Val (IntValue 11)))"),
         new Declaration
             ("f", "x", PARSER.parse("(GaussianKernel (Var x))")));
    public static Program parsedTermImages = new Program
        (PARSER.parse("(App f (Val (IntValue 0)))"),
         new Declaration
             ("f", "x", PARSER.parse("(DisplayImage (LoadImage misc/eindhoven.jpg) Eindhoven)")));
    
    public static void main(String[] args) {
        try {
            Code<Value> cv;
        
            cv = parsedTermDoublePlusInt.peval(tfenv0, fenv0);
            //        System.out.println(cv);
            System.out.println("add1(10.0) = "+cv.run());

            cv = parsedTermIntPlusInt.peval(tfenv0, fenv0);
            //        System.out.println(cv);
            System.out.println("add1(10) = "+cv.run());
        
            cv = parsedTermDoubleEqualsInt.peval(tfenv0, fenv0);
            //        System.out.println(cv);
            System.out.println("equals1(1.0) = "+cv.run());

            cv = parsedTermIntEqualsInt.peval(tfenv0, fenv0);
            //        System.out.println(cv);
            System.out.println("equals1(1) = "+cv.run());
        
            cv = parsedTermDoubleLessInt.peval(tfenv0, fenv0);
            //        System.out.println(cv);
            System.out.println("less1(1.0) = "+cv.run());

            cv = parsedTermIntLessInt.peval(tfenv0, fenv0);
            //        System.out.println(cv);
            System.out.println("less1(1) = "+cv.run());

            cv = parsedTermLet.peval(tfenv0, fenv0);
            //        System.out.println(cv);
            System.out.println("let(123) = "+cv.run());

            cv = parsedTermAllocArray.peval(tfenv0, fenv0);
            //        System.out.println(cv);
            System.out.println("allocArray(10) = "+cv.run());
        
            cv = parsedTermAGet.peval(tfenv0, fenv0);
            //        System.out.println(cv);
            System.out.println("aGet(1) = "+cv.run());

            cv = parsedTermASet.peval(tfenv0, fenv0);
            //        System.out.println(cv);
            System.out.println("aSet(1) = "+cv.run());

            cv = parsedTermASet2.peval(tfenv0, fenv0);
            //        System.out.println(cv);
            System.out.println("aSet2(1) = "+cv.run());

            cv = parsedTermRand.peval(tfenv0, fenv0);
            //        System.out.println(cv);
            System.out.println("rand() = "+cv.run());

            cv = parsedTermFloor.peval(tfenv0, fenv0);
            //        System.out.println(cv);
            System.out.println("floor(3.14) = "+cv.run());

            cv = parsedTermFloorInt.peval(tfenv0, fenv0);
            //        System.out.println(cv);
            System.out.println("floor(3) = "+cv.run());

            cv = parsedTermSFor.peval(tfenv0, fenv0);
            //        System.out.println(cv);
            System.out.println("sFor(10) = "+cv.run());

            cv = parsedTermFor.peval(tfenv0, fenv0);
            //        System.out.println(cv);
            System.out.println("for(10) = "+cv.run());

            cv = parsedTermForNotParallel.peval(tfenv0, fenv0);
            //        System.out.println(cv);
            System.out.println("forNotParallel(10) = "+cv.run());

            cv = parsedTermForNotParallel2.peval(tfenv0, fenv0);
            //        System.out.println(cv);
            System.out.println("forNotParallel2(10) = "+cv.run());
                               
            cv = parsedTermGaussianKernel.peval(tfenv0, fenv0);
            //        System.out.println(cv);
            System.out.println("gaussianKernel(11) = "+cv.run());

            //        cv = parsedTermImages.peval(tfenv0, fenv0);
            ////        System.out.println(cv);
            //        System.out.println("images(0) = "+cv.run());

            // bigger tests
        
            cv = parsedTermFact.peval(tfenv0, fenv0);
            System.out.println(cv);
            System.out.println("fact(10) = "+cv.run());
        
            cv = parsedTermFib.peval(tfenv0, fenv0);
            //        System.out.println(cv);
            System.out.println("fib(10) = "+cv.run());

            pool().shutdown ();
        } finally {
            util.IO.flushOutput ();
        }
    }

}
