package parser;

import java.lang.reflect.*;
import java.util.*;

/**
 * A reflection-based S-expression parser returning an expression of type E.
 * Written by Mathias Ricken and Edwin Westbrook.
 */
public class LintParser<E> {
    /** Convenience method to parse string s into an expression of class c,
      * using the specified subparsers as help, if the parser encounters non-primitive,
      * non-string arguments to S-expressions not within the class hierarcy under class c.
      * @param c class of the expression
      * @param s string to parse
      * @param subParsers subparsers for non-primitive, no-string values
      * @return expression of class c */
    public static <T> T parse(Class<T> c, String s, LintParser<?>... subParsers) {
        LintParser<T> p = new LintParser<T>(c);
        for(LintParser<?> sp: subParsers) {
            p.addSubParser(sp);
        }
        return p.parse(s);
    }
    
    /** Exception being thrown if the string cannot be parsed. */
    public static class ParseException extends RuntimeException {
        public ParseException() { super(); }
        public ParseException(String s) { super(s); }
    }
    
    /** The class of the expression (e.g. lint1_arithmetic.UnstagedLint.Exp). */
    private Class<E> _expClass;
    
    /** The class enclosing the expression classes (e.g. lint1_arithmetic.UnstagedLint). */
    private Class<?> _enclosingClass;
    
    /** A map from S-expression names to classes (e.g. Int -> lint1_arithmetic.UnstagedLint.Int). */
    private HashMap<String,Class<? extends E>> _classes =
        new HashMap<String,Class<? extends E>>();
    
    /** A map from classes for which subparsers have been registered to the subparsers. */
    private HashMap<Class<?>,LintParser<?>> _subParsers =
        new HashMap<Class<?>,LintParser<?>>();
    
    /** A list of exceptions that were thrown. */
    private ArrayList<Throwable> _exceptions = new ArrayList<Throwable>();
    
    /** Create a new parser for expressions of class expClass.
      * @param expClass the class of the expressions to be parsed */
    public LintParser(Class<E> expClass) {
        _expClass = expClass;
        _enclosingClass = _expClass.getEnclosingClass();
        // find all concrete, static subclasses of expClass contained in the same enclosing class
        for(Class<?> c: _enclosingClass.getDeclaredClasses()) {
            if (((c.getModifiers() & Modifier.ABSTRACT)==0) &&
                ((c.getModifiers() & Modifier.STATIC)!=0)) {
                try {
                    Class<? extends E> contained = c.asSubclass(_expClass);
                    _classes.put(contained.getSimpleName(), (Class<? extends E>)contained);
                    println(contained.getSimpleName());
                }
                catch(ClassCastException cce) { /* skip */ }
            }
        }
    }
    
    /** Add a subparser for expressions of type S. Return this parser (not the subparser!)
      * for method chaining.
      * @param p subparser for expressions of type S
      * @return this parser (for expressions of type E) */
    public <S> LintParser<E> addSubParser(LintParser<S> p) {
        _subParsers.put(p._expClass, p);
        return this;
    }    
    
    /** Parse the string.
      * @param s the string to be parsed
      * @return the expression of type E */
    public E parse(String s) {
        println("parse '"+s+"'");
        RestoreStringTokenizer st = new RestoreStringTokenizer(s.trim());
        try {
            return parse(st);
        }
        catch(ParseException e) {
            printExceptions();
            throw e;
        }
    }
    
    /** Parse the tokens in the tokenizer.
      * @param s the tokenizer with the tokens
      * @return the expression of type E */
    protected E parse(RestoreStringTokenizer s) {
        // position to restore everything
        final int sPos = s.getPosition();
        
        // read (
        parseWhitespace(s);
        String word = s.nextToken();
        println("word: '"+word+"'");
        if (!word.equals("(")) {
            s.restorePosition(sPos);
            throwNewParseException("( missing a", s);
        }
        
        // read S-expression name
        parseWhitespace(s);
        word = s.nextToken();
        println("word: '"+word+"'");
        if (!_classes.containsKey(word)) {
            // don't know what that is, restore
            s.restorePosition(sPos);
            throwNewParseException("Don't know "+word, s);
        }
        
        // known subclass of E
        Class<? extends E> c = _classes.get(word);
        println("Class: "+c.getSimpleName());        
        parseWhitespace(s);
        
        // position to restore after trying out a constructor
        final int sTryCtorPos = s.getPosition();
        
        Constructor<? extends E> ctor;
        for(Constructor<?> ctor2: c.getDeclaredConstructors()) {
            println("\ttrying ctor "+ctor2);
            if ((ctor2.getModifiers()&Modifier.PUBLIC)==0) continue; // skip, not public
            
            // this is necessary since c.getDeclaredConstructors() returns Constructor<?>
            // but we want Constructor<? extends E>
            try { ctor = c.getConstructor(ctor2.getParameterTypes()); }
            catch(NoSuchMethodException nsme) { throw new AssertionError("Should never happen: "+nsme); }
            
            // try to parse the arguments for this constructor
            Object[] args = new Object[ctor.getParameterTypes().length];
            int argIndex = 0;
            try {
                for(Class<?> paramC: ctor.getParameterTypes()) {
                    // primitive types and their boxed types
                    if ((paramC==int.class) || (paramC==Integer.class)) {
                        args[argIndex] = parseLeaf(s,Integer.class);
                    }
                    else if ((paramC==boolean.class) || (paramC==Boolean.class)) {
                        args[argIndex] = parseLeaf(s,Boolean.class);
                    }
                    else if ((paramC==long.class) || (paramC==Long.class)) {
                        args[argIndex] = parseLeaf(s,Long.class);
                    }
                    else if ((paramC==double.class) || (paramC==Double.class)) {
                        args[argIndex] = parseLeaf(s,Double.class);
                    }
                    else if ((paramC==float.class) || (paramC==Float.class)) {
                        args[argIndex] = parseLeaf(s,Float.class);
                    }
                    else if ((paramC==byte.class) || (paramC==Byte.class)) {
                        args[argIndex] = parseLeaf(s,Byte.class);
                    }
                    else if ((paramC==short.class) || (paramC==Short.class)) {
                        args[argIndex] = parseLeaf(s,Short.class);
                    }
                    // char or Character
                    else if ((paramC==char.class) || (paramC==Character.class)) {
                        // parse as string
                        Object temp = parseLeaf(s,String.class);
                        // must be exactly one character
                        if (temp.toString().length()!=1) {
                            throwNewParseException("char or Character expected, but got string of length "+
                                                   temp.toString().length()+": "+temp.toString(), s);
                        }
                        args[argIndex] = new Character(temp.toString().charAt(0));
                    }                    
                    // strings
                    else if (paramC==String.class) {
                        args[argIndex] = ((String)parseLeaf(s,String.class)).intern(); // intern all strings
                    }
                    // recursively parse expressions of type E
                    else if (_expClass.equals(paramC)) {
                        args[argIndex] = parse(s);
                    }
                    // try to use a subparser
                    else if (_subParsers.containsKey(paramC)) {
                        // try one of the subparsers
                        args[argIndex] = _subParsers.get(paramC).parse(s);
                    }
                    else {
                        // don't know what that is
                        // restore happens below
                        throwNewParseException("Don't know how to parse "+paramC, s);
                    }
                    ++argIndex;
                }
            }
            catch(ParseException pe) {
                println(pe);
                // this constructor didn't work out, we need to restore and try another constructor
                s.restorePosition(sTryCtorPos);
                // continue with next constructor
                continue;
            }
            try {
                // we read all values required for this constructor
                // make sure that the next token is )
                parseWhitespace(s);
                word = s.nextToken();
                println("word: '"+word+"'");
                if (!word.equals(")")) {
                    // it wasn't ), we need to restore and try another constructor
                    s.restorePosition(sTryCtorPos);
                    // continue with next constructor
                    continue;
                }
                parseWhitespace(s);
                
                // successfully used this constructor
                println("new "+ctor.getDeclaringClass().getSimpleName()+" "+
                        java.util.Arrays.toString(args));
                E e = (E)ctor.newInstance(args);
                return e;
            }
            catch(Exception e) {
                println(e);
                // something went wrong using this constructor, we need to restore and try another
                s.restorePosition(sTryCtorPos);
                // continue with next constructor
                continue;
            }
        }
        
        s.restorePosition(sPos);
        throwNewParseException("Could not find a constructor for "+c, s); // always throws
        return null; // cannot be reached
    }
    
    /** Parse a leaf of class c, which must have a unary constructor taking a String, from
      * the tokenizer s.
      * @param s tokenizer from which to parse
      * @param c the class, which has a unary constructor taking a String, for that we want to create a value
      * @return parsed value of class c */
    protected Object parseLeaf(RestoreStringTokenizer s, Class<?> c) {
        // position so we can undo this attempt to parse
        final int sPos = s.getPosition();
        
        String word = null;
        try {
            // get the unary constructor taking String
            Constructor ctor = c.getDeclaredConstructor(String.class);
            
            // skip whitespace, get the string, skip whitespace
            parseWhitespace(s);
            word = s.nextToken();
            println("word: '"+word+"'");
            Object o = ctor.newInstance(word);
            parseWhitespace(s);
            
            return o;
        }
        catch(NoSuchMethodException nsme) {
            // something went wrong, restore and abort
            s.restorePosition(sPos);
            throwNewParseException("Could not create leaf of "+c, s); 
        }
        catch(InstantiationException ie) {
            // something went wrong, restore and abort
            s.restorePosition(sPos);
            throwNewParseException("Could not create leaf of "+c, s); 
        }
        catch(IllegalAccessException iae) {
            // something went wrong, restore and abort
            s.restorePosition(sPos);
            throwNewParseException("Could not create leaf of "+c, s); 
        }
        catch(InvocationTargetException ite) {
            // something went wrong, restore and abort
            s.restorePosition(sPos);
            throwNewParseException("Could not create leaf of "+c, s); 
        }
        return null; // cannot be reached
    }
    
    /** Parse whitespace. Stop before the next non-whitespace token, or when there are no
      * more tokens.
      * @param s tokenizer from which to parse whitespace */
    protected void parseWhitespace(RestoreStringTokenizer s) {
        String word;
        int sPrevPos;
        while(s.hasMoreTokens()) {
            sPrevPos = s.getPosition();
            word = s.nextToken();
            if (!word.trim().equals("")) {
                s.restorePosition(sPrevPos);
                break;
            }
        }
    }
    
    /** Add a new ParseException to the list of exceptions and throw it. */
    protected void throwNewParseException(RestoreStringTokenizer s) {
        String allTokensRead = s.allTokensRead();
        if (allTokensRead.length()>40) {
            allTokensRead = "... " + allTokensRead.substring(allTokensRead.length()-36, allTokensRead.length());
        }
        if (allTokensRead.length()>0) {
            throwParseException(new ParseException("At: "+allTokensRead));
        }
        else {
            throwParseException(new ParseException());
        }
    }
    
    /** Add a new ParseException to the list of exceptions and throw it. */
    protected void throwNewParseException(String msg, RestoreStringTokenizer s) {
        String allTokensRead = s.allTokensRead();
        if (allTokensRead.length()>40) {
            allTokensRead = "... " + allTokensRead.substring(allTokensRead.length()-36, allTokensRead.length());
        }
        if (allTokensRead.length()>0) {
            throwParseException(new ParseException(msg+", at: "+allTokensRead));
        }
        else {
            throwParseException(new ParseException(msg));
        }
    }
    
    /** Add a new ParseException to the list of exceptions and throw it. */
    protected void throwParseException(ParseException e) {
        _exceptions.add(e);
        throw e;
    }
    
    /** Return the list of exceptions thrown.
      * @return list of exceptions */
    public ArrayList<Throwable> getExceptions() {
        return _exceptions;
    }
    
    /** Print exceptions. */
    public void printExceptions() {
        for(Throwable t: _exceptions) {
            // System.err.println(t);
        }
    }
    
    /** Debug method to print, comment System.out.println call out to disable debug printing */
    public static void println(Object o) {
        // System.out.println(o.toString());
    }
    
    /** This tokenizer stores all read tokens in a stack so they can easily be
      * restored. This isn't as memory as it could be, but it's good enough for
      * small examples. */
    public static class RestoreStringTokenizer extends StringTokenizer {
        /** Stack of read tokens. */
        private Stack<String> _read = new Stack<String>();
        /** Stack of restored tokens. */
        private Stack<String> _restored = new Stack<String>();
        /** Create a new string tokenizer that can restore positions. */
        public RestoreStringTokenizer(String s) { super(s," ()",true); }
        /** Return the current position. */
        public int getPosition() { return _read.size(); }
        /** Rewind to a previous position. This only goes backward, not forward. */
        public void restorePosition(int pos) {
            while(_read.size()>pos) {
                _restored.push(_read.pop());
            }
        }
        public String allTokensRead() {
            StringBuilder sb = new StringBuilder();
            for(String t: _read) { sb.append(t).append(' ');; }
            return sb.toString().trim();
        }
        // overridden methods from StringTokenizer
        public int countTokens() { return _restored.size() + super.countTokens(); }
        public boolean hasMoreTokens() { return (_restored.size()>0) || super.hasMoreTokens(); }
        public boolean hasMoreElements() { return hasMoreTokens(); }
        public String nextToken() {
            String token = (_restored.size()>0) ? _restored.pop() : super.nextToken();
            _read.push(token);
            return token;
        }
        public Object nextElement() { return nextToken(); }
        public String nextToken(String delim) {
            if (_restored.size()>0) return _restored.pop();
            return super.nextToken(delim);
        }
    }
}
