package simple_int;

import edu.rice.cs.mint.runtime.Code;
import edu.rice.cs.mint.runtime.MintSerializer;

import java.io.*;

public class StagedLint {
    public static interface Exp {
        public separable Code<Integer> eval(Env e, FEnv f);
    }
    
    public static class Int implements Exp {
        private Code<Integer> _value;
        public separable Int(final int value) {
            _value = <| value |>;
        }
        public separable Int(Code<Integer> value) {
            _value = value;
        }
        public separable Code<Integer> eval(Env e, FEnv f) {
            return _value;
        }
    }
    
    public static class Var implements Exp {
        private String _s;
        public separable Var(String s) {
            _s = s; // s must have been interned, so we can use == in the fext method below
        }
        public separable Code<Integer> eval(Env e, FEnv f) {
            return e.get(_s);
        }
    }
    
    public static class App implements Exp {
        private String _s;
        private Exp _arg;
        public separable App(String s, Exp arg) {
            _s = s; // s must have been interned, so we can use == in the fext method above
            _arg = arg;
        }
        public separable Code<Integer> eval(Env e, FEnv f) {
            return <| `(f.get(_s)).apply(`(_arg.eval(e,f))) |>;
        }
    }
    
    public static abstract class BinOp implements Exp {
        protected Exp _left, _right;
        public separable BinOp(Exp left, Exp right) {
            _left = left;
            _right = right;
        }
    }
    
    public static class Add extends BinOp {
        public separable Add(Exp left, Exp right) { super(left, right); }
        public separable Code<Integer> eval(Env e, FEnv f) {
            return <| `(_left.eval(e,f)) + `(_right.eval(e,f)) |>;
        }
    }
    
    public static class Sub extends BinOp {
        public separable Sub(Exp left, Exp right) { super(left, right); }
        public separable Code<Integer> eval(Env e, FEnv f) {
            return <| `(_left.eval(e,f)) - `(_right.eval(e,f)) |>;
        }
    }
    
    public static class Mul extends BinOp {
        public separable Mul(Exp left, Exp right) { super(left, right); }
        public separable Code<Integer> eval(Env e, FEnv f) {
            return <| `(_left.eval(e,f)) * `(_right.eval(e,f)) |>;
        }
    }
    
    public static class Div extends BinOp {
        public separable Div(Exp left, Exp right) { super(left, right); }
        public separable Code<Integer> eval(Env e, FEnv f) {
            return <| `(_left.eval(e,f)) / `(_right.eval(e,f)) |>;
        }
    }
    
    public static class Ifz implements Exp {
        private Exp _test, _conseq, _alt;
        public separable Ifz(Exp test, Exp conseq, Exp alt) {
            _test = test;
            _conseq = conseq;
            _alt = alt;
        }
        public separable Code<Integer> eval(Env e, FEnv f) {
            return <| ((`(_test.eval(e,f))==0)?
                           `(_conseq.eval(e,f)):
                           `(_alt.eval(e,f))) |>;
        }
    }
    
    public static class EnvironmentException extends RuntimeException { }
    
    // interface to represent int -> int functions
    public static interface IntFun {
        public int apply(int param);
    }
    
    // environment; implemented as a function object from String to Code<Integer>
    public static interface Env {
        public separable Code<Integer> get(String y);
    }
    public static final Env env0 = new Env() {
        public separable Code<Integer> get(String s) {
            throw new EnvironmentException();
        }
    };
    
    public static separable Env ext(final Env env,
                                    final String x,
                                    final Code<Integer> v) {
        return new Env() {
            public separable Code<Integer> get(String y) {
                // error: if (x.equals(y))
                if (x==y)
                    return v;
                else
                    return env.get(y);
            }
        };
    }     
    
    // function environment; implemented as a function object from String to Code<IntFun>
    public static interface FEnv {
        public separable Code<? extends IntFun> get(String s);
    }
    public static final FEnv fenv0 = new FEnv() {
        public separable Code<? extends IntFun> get(String s) {
            throw new EnvironmentException();
        }
    };
    
    public static separable FEnv fext(final FEnv fenv, final String x, final Code<? extends IntFun> v) {
        return new FEnv() {
            public separable Code<? extends IntFun> get(String y) {
                // error: if (x.equals(y))
                if (x==y)
                    return v;
                else
                    return fenv.get(y);
            }
        };
    }
    
    public static class Declaration {
        private String _fun, _param;
        private Exp _body;
        public separable Declaration(String fun, String param, Exp body) {
            _fun = fun; // fun must have been interned, so we can use == in the fext method above
            _param = param; // param must have been interned, so we can use == in the fext method above
            _body = body;
        }
        public separable String fun() { return _fun; }
        public separable String param() { return _param; }
        public separable Exp body() { return _body; }
    }
    
    public static class Program {
        private Declaration[] _decls;
        private Exp _body;
        public separable Program(Exp body, Declaration ... decls) {
            _decls = decls;
            _body = body;
        }
        
        public separable Code<Integer> peval(Env env, FEnv fenv) {
            return peval(env,fenv,0);
        }
        
        private separable Code<Integer> peval(final Env env, final FEnv fenv, int declIndex) {
            // match p with
            if (_decls.length<=declIndex) {
                //     Program ([],e) -> eval e env fenv
                return _body.eval(env,fenv);
            }
            else {
                //    |Program (Declaration (s1,s2,e1)::tl,e) ->
                //        let rec f x = eval e1 (ext env s2 x) (ext fenv s1 f)
                //        in peval (Program(tl,e)) env (ext fenv s1 f)
                final Declaration d = _decls[declIndex];
                final String dParam = d.param();
                final String dFun = d.fun();
                final Exp dBody = d.body();
                return peval(env,fext(fenv, dFun, <|new IntFun() {
                    public int apply(final int x) {
                        final IntFun fthis = this;
                        return `(dBody.eval(ext(env, dParam, <| x |>),
                                            fext(fenv, dFun, <| fthis |>)));
                    }
                }|>),declIndex+1);
            }
        }
    }
    
    public static Program termFact = new Program
        (new App("f", new Int(10)),
         new Declaration
             ("f", "x", new Ifz
                  (new Var("x"),
                   new Int(1),
                   new Mul(new Var("x"),
                           new App
                               ("f",
                                new Sub
                                    (new Var("x"),
                                     new Int(1)))))));

    public static Program parsedTermFact = new Program
        (parser.LintParser.parse(Exp.class, "(App f (Int 10))"),
         new Declaration
             ("f", "x", parser.LintParser.parse
                  (Exp.class, "(Ifz (Var x) (Int 1) (Mul (Var x) (App f (Sub (Var x) (Int 1)))))")));
    
    public static Program termFib = new Program
        (new App("f", new Int(10)),
         new Declaration("f", "x", new Ifz
                             (new Var("x"),
                              new Int(0),
                              new Ifz
                                  (new Sub
                                       (new Var("x"),
                                        new Int(1)),
                                   new Int(1),
                                   new Add
                                       (new App
                                            ("f",
                                             new Sub
                                                 (new Var("x"),
                                                  new Int(1))),
                                        new App
                                            ("f",
                                             new Sub
                                                 (new Var("x"),
                                                  new Int(2))))))));

    public static Program parsedTermFib = new Program
        (parser.LintParser.parse(Exp.class, "(App f (Int 10))"),
         new Declaration
             ("f", "x", parser.LintParser.parse
                  (Exp.class, "(Ifz (Var x) (Int 0) (Ifz (Sub (Var x) (Int 1)) (Int 1) "+
                   "(Add (App f (Sub (Var x) (Int 1))) (App f (Sub (Var x) (Int 2)))))))")));
    
    public static separable Code<Integer> bindParameter(Exp functionBody, Code<Integer> param) {
        return new Program
            (new App("f", new Int(param)),
             new Declaration
                 ("f", "n", functionBody)).peval(env0, fenv0);
    }
    
    public static void main(String[] args) throws IOException {
        Code<IntFun> codeFun = <| (IntFun) new IntFun() {
            public int apply(final int x) {
                return `(new Add(new Int(10), new Int(<|x|>)).eval(env0, fenv0));
            }
        } |>;

        
        Code<Integer> c = termFact.peval(env0, fenv0);
        System.out.println(c);
        System.out.println("fact(10) = "+c.run());
        
        c = parsedTermFact.peval(env0, fenv0);
        System.out.println(c);
        System.out.println("fact(10) = "+c.run());
        
        c = termFib.peval(env0, fenv0);
        System.out.println(c);
        System.out.println("fib(10) = "+c.run());
        
        c = parsedTermFib.peval(env0, fenv0);
        System.out.println(c);
        System.out.println("fib(10) = "+c.run());
        
        
        File dir = new File("tempClasses");
        dir.mkdir();
        
        // Fibonacci
        final Exp fibBody = parser.LintParser.parse
            (Exp.class, "(Ifz (Var x) (Int 0) (Ifz (Sub (Var x) (Int 1)) (Int 1) "+
             "(Add (App f (Sub (Var x) (Int 1))) (App f (Sub (Var x) (Int 2)))))))");
        Code<? extends IntFun> codeFibFun = <| new IntFun() {
            public int apply(final int param) {
                return `(bindParameter(fibBody, <|param|>));
            }
        } |>;

        System.out.println(codeFibFun);

        File fibFile = new File(dir, "FibFun.jar");
        MintSerializer.save(codeFibFun, "FibFun", fibFile);

        IntFun sFibFun = codeFibFun.run();
        System.out.println("sFibFun(10) = "+sFibFun.apply(10));
        System.out.println("sFibFun(12) = "+sFibFun.apply(12));
        System.out.println("sFibFun(15) = "+sFibFun.apply(15));
        
        StagedLint.IntFun sFibFun2 = MintSerializer.load(fibFile);
        System.out.println("sFibFun2(10) = "+sFibFun2.apply(10));
        System.out.println("sFibFun2(12) = "+sFibFun2.apply(12));
        System.out.println("sFibFun2(15) = "+sFibFun2.apply(15));

        // Factorial
        final Exp factBody = parser.LintParser.parse
            (Exp.class, "(Ifz (Var x) (Int 1) (Mul (Var x) (App f (Sub (Var x) (Int 1)))))");
        
        Code<? extends IntFun> codeFactFun = <| new IntFun() {
            public int apply(final int param) {
                return `(bindParameter(factBody, <|param|>));
            }
        } |>;
        
        System.out.println(codeFactFun);
        
        File factFile = new File(dir, "FactFun.jar");
        MintSerializer.save(codeFactFun, "FactFun", factFile);
        
        IntFun sFactFun = codeFactFun.run();
        System.out.println("sFactFun(10) = "+sFactFun.apply(10));
        System.out.println("sFactFun(12) = "+sFactFun.apply(12));
        System.out.println("sFactFun(15) = "+sFactFun.apply(15));
        
        StagedLint.IntFun sFactFun2 = MintSerializer.load(factFile);
        System.out.println("sFactFun2(10) = "+sFactFun2.apply(10));
        System.out.println("sFactFun2(12) = "+sFactFun2.apply(12));
        System.out.println("sFactFun2(15) = "+sFactFun2.apply(15));
    }
}
